# -*- coding: utf-8 -*-
"""
orm query base module.
"""

import inspect

from sqlalchemy import inspection, log, func, literal
from sqlalchemy.orm import Query, lazyload
from sqlalchemy.orm.attributes import InstrumentedAttribute

import pyrin.utils.misc as misc_utils
import pyrin.database.paging.services as paging_services

from pyrin.core.globals import _
from pyrin.database.model.base import BaseEntity
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.services import get_current_store
from pyrin.database.orm.query.exceptions import ColumnsOutOfScopeError, \
    UnsupportedQueryStyleError


@inspection._self_inspects
@log.class_logger
class CoreQuery(Query):
    """
    core query class.

    this is the application default query class.
    it extends sqlalchemy `Query` class.
    """

    def __init__(self, entities, session=None, **options):
        """
        initializes an instance of CoreQuery.

        this method has been overridden to provide the concept of scope to queries.
        it is useful if you want to let users (end users not developers) to select
        which columns they want to be returned in a service. in this situation, if
        there is no scope defined, they could add any columns of other entities, but
        using scope, prevents this. but on normal use cases there is no need to define
        scope, and its also more efficient.

        :param tuple entities: entities or columns that are needed for query.
        :param Session session: optional session object to bind this query to it.

        :keyword type | tuple[type] scope: class type of the entities that this
                                           query instance will work on. if the
                                           query is working on multiple entities,
                                           this value must be a tuple of all class
                                           types of that entities.

                                           for example: if you set
                                           `entities=SomeEntity.id, AnotherEntity.name`
                                           you should leave `scope=None` to skip validation
                                           or you could set
                                           `scope=(SomeEntity, AnotherEntity)`
                                           this way validation succeeds, but if
                                           you set `scope=SomeEntity`
                                           then the query will not be executed
                                           and an error will be raised.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope = options.get('scope', None)
        if scope is not None:
            self._prepare_validation(entities, scope)

        super().__init__(entities, session)

    def _validate_scope(self, entities, scope):
        """
        validates the query for given entities based on provided scope.

        :param tuple entities: entities or columns that needed for query.

        :param type[BaseEntity] | tuple[type[BaseEntity]] scope: class type of the entities
                                                                 that this query instance will
                                                                 work on. if the query is working
                                                                 on multiple entities, this value
                                                                 must be a tuple of all class
                                                                 types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope_classes = set(entity for entity in scope if inspect.isclass(entity)
                            and issubclass(entity, BaseEntity))

        requested_classes = set(entity for entity in entities if inspect.isclass(entity)
                                and issubclass(entity, BaseEntity))

        requested_classes_by_column = set(column.class_ for column in entities if
                                          isinstance(column, InstrumentedAttribute)
                                          and issubclass(getattr(column, 'class_', type),
                                                         BaseEntity))

        all_requested_classes = requested_classes.union(requested_classes_by_column)

        if not all_requested_classes.issubset(scope_classes):
            raise ColumnsOutOfScopeError(_('Requested columns are out of scope of this '
                                           'query. Please revise the requested columns.'))

    def _prepare_validation(self, entities, scope):
        """
        prepares entities and scope and validates them if required.

        :param tuple entities: entities or columns that needed for query.

        :param type[BaseEntity] | tuple[type[BaseEntity]] scope: class type of the entities that
                                                                 this query instance will work on.
                                                                 if the query is working on
                                                                 multiple entities, this value
                                                                 must be a tuple of all class
                                                                 types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        if scope is None or entities is None:
            return

        scope = misc_utils.make_iterable(scope, tuple)
        entities = misc_utils.make_iterable(entities, tuple)

        if len(scope) <= 0 or len(entities) <= 0:
            return

        self._validate_scope(entities, scope)

    def count(self, **options):
        """
        returns the count of rows that the sql formed by this `Query` would return.

        this method is overridden to prevent inefficient count() of sqlalchemy `Query`
        which produces a subquery.

        this method generates a single sql query like below:
        select count(column, ...)
        from table
        where ...

        :keyword bool distinct: specifies that count should
                                be executed on distinct select.
                                defaults to False if not provided.

        :keyword bool fallback: specifies that if the overridden count
                                failed to execute, it should be executed using
                                the original sqlalchemy count which produces
                                a subquery, instead of raising an error.
                                defaults to False if not provided.

        :rtype: int
        """

        fallback = options.get('fallback', False)
        needs_fallback = False
        columns = []
        # if there is group by clause, a subquery
        # is inevitable to be able to get count.
        if self.selectable._group_by_clause is not None and \
                self.selectable._group_by_clause.clauses is not None and \
                len(self.selectable._group_by_clause.clauses) > 0:
            needs_fallback = True
        else:
            for single_column in self.selectable.columns:
                if not isinstance(single_column, CoreColumn):
                    if fallback is False:
                        raise UnsupportedQueryStyleError('Current query does not have columns '
                                                         'of type [{column_type}] in its '
                                                         'expression. if you need to apply a '
                                                         '"DISTINCT" keyword, you should apply '
                                                         'it by passing "distinct=True" keyword '
                                                         'to count() method and do not apply it '
                                                         'in query structure itself. for example '
                                                         'instead of writing "store.query('
                                                         'distinct(Entity.id)).count()" you '
                                                         'should write this in the following '
                                                         'form "store.query(Entity.id).count('
                                                         'distinct=True)". but if you want the '
                                                         'sqlalchemy original style of count() '
                                                         'which produces a subquery, it is also '
                                                         'possible to fallback to that default '
                                                         'sqlalchemy count() but keep in mind '
                                                         'that, that method is not efficient. '
                                                         'you could pass "fallback=True" in '
                                                         'options to fallback to default mode '
                                                         'if overridden count() method failed '
                                                         'to provide count.'
                                                         .format(column_type=CoreColumn))
                    else:
                        needs_fallback = True
                        break

                fullname = single_column.fullname
                if fullname not in (None, ''):
                    columns.append(fullname)

        if needs_fallback is True:
            return super().count()

        func_count = func.count()
        if len(columns) > 0:
            distinct = options.get('distinct', False)
            column_clause = ', '.join(columns)
            if distinct is True:
                column_clause = 'distinct {clause}'.format(clause=column_clause)
            func_count = func.count(column_clause)

        statement = self.options(lazyload('*')).statement.with_only_columns(
            [func_count]).order_by(None)

        store = get_current_store()
        return store.execute(statement).scalar()

    def paginate(self, **options):
        """
        sets offset and limit for current query.

        the offset and limit values will be extracted from given inputs.

        :keyword int __limit__: limit value.
        :keyword int __offset__: offset value.
        """

        limit, offset = paging_services.get_paging_keys(**options)
        return self.limit(limit).offset(offset)

    def existed(self):
        """
        gets a value indicating that current query has any results.

        this is a helper method to simplify the use of `.exists` method of sqlalchemy.

        :rtype: bool
        """

        store = get_current_store()
        result = store.query(literal(True)).filter(self.exists()).scalar()
        return result is True

    def delete(self, synchronize_session=False):
        """
        performs a bulk delete query.

        this method is overridden to provide the most performant
        `synchronize_session` value as default value.

        :param str | bool synchronize_session: session synchronization strategy.
                                               it could be set to False, `fetch`
                                               or `evaluate`. defaults to False.

        :returns: count of affected rows
        :rtype: int
        """

        return super().delete(synchronize_session=synchronize_session)

    def update(self, values, synchronize_session=False, update_args=None):
        """
        performs a bulk update query.

        this method is overridden to provide the most performant
        `synchronize_session` value as default value.

        :param values: a dictionary with attributes names, or alternatively
                       mapped attributes or sql expressions, as keys, and
                       literal values or sql expressions as values.

        :param str | bool synchronize_session: session synchronization strategy.
                                               it could be set to False, `fetch`
                                               or `evaluate`. defaults to False.

        :param update_args: optional dictionary, if present will be passed
                            to the underlying `_expression.update` construct
                            as the `**kw` for the object.

        :returns: count of affected rows
        :rtype: int
        """

        return super().update(values,
                              synchronize_session=synchronize_session,
                              update_args=update_args)



# -*- coding: utf-8 -*-
"""
celery cli interface module.
"""

import pyrin.task_queues.celery.cli.services as celery_cli_services

from pyrin.cli.base import CLIHandlerBase
from pyrin.cli.params import CLIParamBase
from pyrin.core.globals import LIST_TYPES


class CeleryCLIParamBase(CLIParamBase):
    """
    celery cli param base class.

    all celery cli param classes must be subclassed from this.
    """

    def _convert_result(self, value):
        """
        converts the given value to required type.

        celery command line arguments must be all strings.

        :param list[object] | object value: value to be converted.

        :rtype: list[object] | object
        """

        if value is not None and not isinstance(value, str):
            if isinstance(value, LIST_TYPES):
                original_type = type(value)
                converted_results = []
                for item in value:
                    if item is not None and not isinstance(item, str):
                        converted_results.append(str(item))
                    else:
                        converted_results.append(item)
                return original_type(converted_results)

            else:
                return str(value)

        return value


class CeleryCLIHandlerBase(CLIHandlerBase):
    """
    celery cli handler base class.

    all celery cli handlers must be subclassed from this.
    """

    def __init__(self, name):
        """
        initializes an instance of CeleryCLIHandlerBase.

        :param str name: the handler name that should be registered
                         with. this name must be the exact name that
                         this handler must emmit to cli.
        """

        super().__init__(name)

    def _get_common_cli_options(self):
        """
        gets the list of common cli options.

        :rtype: list
        """

        return ['celery']

    def _inject_command_name(self, commands):
        """
        injects the command name into given commands list.

        this method is overridden to prevent adding the command name into commands list.

        :param list commands: a list of all commands to inject
                              this command's name into it.
        """
        pass

    def _execute_on_cli(self, commands):
        """
        executes the current cli command with given inputs.

        :param list commands: a list of all commands and their
                              values to be sent to cli command.

        :rtype: int
        """

        result = celery_cli_services.execute_on_celery(self.get_name(), commands)
        return self._process_return_value(result)



# -*- coding: utf-8 -*-
"""
validator interface module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject


class AbstractValidatorBase(CoreObject):
    """
    abstract validator base class.

    all application validators must be subclassed from this.
    """

    @abstractmethod
    def validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.

        :param object value: value to be validated.

        :raises CoreNotImplementedError: core not implemented error.
        :raises ValidationError: validation error.
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        gets the name of this validator.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def domain(self):
        """
        gets the domain of this validator.

        domain is the type of a BaseEntity subclass that
        this validator validates a value of it.
        if a validator is not specific to an entity, then
        domain could be a unique string name.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type[BaseEntity] | str
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
permission manager module.
"""

from abc import abstractmethod

import pyrin.database.bulk.services as bulk_services

from pyrin.core.globals import SECURE_FALSE
from pyrin.core.structs import Context, Manager
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.permission import PermissionPackage
from pyrin.security.permission.base import PermissionBase
from pyrin.security.permission.exceptions import InvalidPermissionTypeError, \
    DuplicatedPermissionError


class PermissionManager(Manager):
    """
    permission manager class.
    """

    package_class = PermissionPackage

    def __init__(self):
        """
        initializes an instance of PermissionManager.
        """

        super().__init__()

        # holds a dict of all application's loaded permissions.
        # in the form of dict[object permission_id: PermissionBase instance]
        self.__permissions = Context()

    def register_permission(self, instance, **options):
        """
        registers the given permission.

        :param PermissionBase instance: permission instance to be registered.

        :raises InvalidPermissionTypeError: invalid permission type error.
        :raises DuplicatedPermissionError: duplicated permission error.
        """

        if not isinstance(instance, PermissionBase):
            raise InvalidPermissionTypeError('Input parameter [{instance}] is '
                                             'not an instance of [{base}].'
                                             .format(instance=instance,
                                                     base=PermissionBase))

        if instance.get_id() in self.__permissions:
            raise DuplicatedPermissionError('Permission [{permission}] has been '
                                            'already registered.'
                                            .format(permission=instance))

        self.__permissions[instance.get_id()] = instance

    def get_permissions(self, **options):
        """
        gets all registered permissions.

        :rtype: list[PermissionBase]
        """

        return self.__permissions.values()

    def synchronize_all(self, **options):
        """
        synchronizes all permissions with database.

        it creates or updates the available permissions.
        """

        entities = [permission.to_entity() for permission in self.get_permissions()]
        needs_update = [entity for entity in entities if
                        self._exists(*entity.primary_key(as_tuple=True)) is True]
        needs_insert = list(set(entities).difference(set(needs_update)))

        if needs_insert:
            bulk_services.insert(*needs_insert, readable=SECURE_FALSE)
        if needs_update:
            bulk_services.update(*needs_update, readable=SECURE_FALSE)

    @abstractmethod
    def _exists(self, *primary_key):
        """
        gets a value indicating that given permission exists in database.

        this method must be implemented in subclasses, the input value
        could be as many as needed arguments to represent the primary key
        of your permission entity.
        if you don't want to use permissions concept in your application,
        you could leave this method unimplemented.

        :param object primary_key: permission primary key value.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
city exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CityManagerException(CoreException):
    """
    city manager exception.
    """
    pass


class CityManagerBusinessException(CoreBusinessException,
                                   CityManagerException):
    """
    city manager business exception.
    """
    pass


class CityNotFoundError(CityManagerBusinessException):
    """
    city not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
permission component module.
"""

from pyrin.application.structs import Component
from pyrin.application.decorators import component

from demo.security.permission import PermissionPackage
from demo.security.permission.manager import PermissionManager


@component(PermissionPackage.COMPONENT_NAME, replace=True)
class PermissionComponent(Component, PermissionManager):
    """
    permission component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
audit hooks module.
"""

import pyrin.audit.services as audit_services
import pyrin.configuration.services as config_services

from pyrin.core.structs import Hook
from pyrin.utils.custom_print import print_info
from pyrin.application.decorators import application_hook
from pyrin.application.hooks import ApplicationHookBase


class AuditHookBase(Hook):
    """
    audit hook base class.

    all packages that need to be hooked into audit business must
    implement this class and register it in audit hooks.
    """

    # name of the current audit, for example: database or celery.
    # the result of each audit will be added to final result dict using this key.
    audit_name = None

    def inspect(self, **options):
        """
        this method will be called to inspect the status of a package or resource.

        each subclass must return a tuple of two values. first value is a dict
        containing the inspection data. and the second value is a bool value
        indicating that inspection has been succeeded or failed.

        subclasses could also raise an exception instead of returning a value to
        indicate the failure, if required.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :rtype: tuple[dict, bool]
        """

        return {}, True


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def before_application_run(self):
        """
        this method will be get called just before application gets running.

        note that this method will not get called when
        application starts in scripting mode.

        :raises AuditFailedError: audit failed error.
        """

        startup_audit = config_services.get_active('audit', 'startup_audit')
        if startup_audit is True:
            print_info('Performing audit...')
            audit_services.startup_inspect()



# -*- coding: utf-8 -*-
"""
audit package.
"""

from pyrin.packaging.base import Package


class AuditPackage(Package):
    """
    audit package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.api.router',
               'pyrin.configuration']
    COMPONENT_NAME = 'audit.component'
    CONFIG_STORE_NAMES = ['audit']



# -*- coding: utf-8 -*-
"""
datetime package.
"""

from pyrin.packaging.base import Package


class DateTimePackage(Package):
    """
    datetime package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'globalization.datetime.component'
    DEPENDS = ['pyrin.configuration',
               'pyrin.security.session']



# -*- coding: utf-8 -*-
"""
users services module.
"""

from pyrin.application.services import get_component
from pyrin.security.users import UsersPackage


def is_active(user, **options):
    """
    gets a value indicating that given user is active.

    :param user: user to check its active status.

    :rtype: bool
    """

    return get_component(UsersPackage.COMPONENT_NAME).is_active(user, **options)



# -*- coding: utf-8 -*-
"""
audit exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class AuditManagerException(CoreException):
    """
    audit manager exception.
    """
    pass


class AuditManagerBusinessException(CoreBusinessException, AuditManagerException):
    """
    audit manager business exception.
    """
    pass


class InvalidAuditHookTypeError(AuditManagerException):
    """
    invalid audit hook type error.
    """
    pass


class AuditFailedError(AuditManagerException):
    """
    audit failed error.
    """
    pass



# -*- coding: utf-8 -*-
"""
babel enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class BabelCLIHandlersEnum(CoreEnum):
    """
    babel cli handlers enum.
    """

    COMPILE = 'compile'
    EXTRACT = 'extract'
    INIT = 'init'
    UPDATE = 'update'



# -*- coding: utf-8 -*-
"""
router handlers package.
"""

from pyrin.packaging.base import Package


class RouterHandlersPackage(Package):
    """
    router handlers package class.
    """

    NAME = __name__
    DEPENDS = []



# -*- coding: utf-8 -*-
"""
province manager module.
"""

from pyrin.core.structs import Manager
from pyrin.database.services import get_current_store
from pyrin.core.globals import _

from demo.common.province.exceptions import ProvinceNotFoundError
from demo.common.province.models import ProvinceEntity


class ProvinceManager(Manager):
    """
    province manager class.
    """

    def get(self, province_id, **options):
        """
        gets the specified province.

        :param int province_id: province id to get its info.

        :raises ProvinceNotFoundError: province not found error.

        :returns: dict(int id,
                       str name)

        :rtype: dict
        """

        store = get_current_store()
        province = store.query(ProvinceEntity).get(province_id)

        if province is None:
            raise ProvinceNotFoundError(_('Province [{province_id}] not found.'
                                          .format(province_id=province_id)))

        return province.to_dict()

    def find(self, **filters):
        """
        finds provinces with given filters.

        :keyword str name: province name.

        :returns: list[dict(int id,
                            str name)]

        :rtype: list
        """

        clauses = self._make_find_clause(**filters)

        store = get_current_store()
        entities = store.query(ProvinceEntity).filter(*clauses).all()

        return entities

    def get_all(self, **options):
        """
        gets all provinces.

        :returns: list[dict(int id,
                            str name)]

        :rtype: list
        """

        return self.find()

    def _make_find_clause(self, **filters):
        """
        makes the required find clauses based on given
        filters and returns the clauses list.

        :keyword str name: province name.

        :rtype: list
        """

        clauses = []

        name = filters.get('name', None)

        if name is not None:
            clauses.append(ProvinceEntity.name.icontains(name))

        return clauses

    def create(self, name, **options):
        """
        creates a province.

        :param str name: province name.
        """

        province = ProvinceEntity(name=name)
        store = get_current_store()
        store.add(province)



# -*- coding: utf-8 -*-
"""
locale component module.
"""

from pyrin.application.decorators import component
from pyrin.globalization.locale import LocalePackage
from pyrin.globalization.locale.manager import LocaleManager
from pyrin.application.structs import Component


@component(LocalePackage.COMPONENT_NAME)
class LocaleComponent(Component, LocaleManager):
    """
    locale component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
province services module.
"""

from pyrin.application.services import get_component

from demo.common.province import ProvincePackage


def get(province_id, **options):
    """
    gets the specified province.

    :param int province_id: province id to get its info.

    :raises ProvinceNotFoundError: province not found error.

    :returns: dict(int id,
                   str name)

    :rtype: dict
    """

    return get_component(ProvincePackage.COMPONENT_NAME).get(province_id, **options)


def find(**filters):
    """
    finds provinces with given filters.

    :keyword str name: province name.

    :returns: list[dict(int id,
                        str name)]

    :rtype: list
    """

    return get_component(ProvincePackage.COMPONENT_NAME).find(**filters)


def get_all(**options):
    """
    gets all provinces.

    :returns: list[dict(int id,
                        str name)]

    :rtype: list
    """

    return get_component(ProvincePackage.COMPONENT_NAME).get_all(**options)


def create(name, **options):
    """
    creates a province.

    :param str name: province name.
    """

    return get_component(ProvincePackage.COMPONENT_NAME).create(name, **options)



# -*- coding: utf-8 -*-
"""
cors enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class CORSRequestHeaderEnum(CoreEnum):
    """
    cors request header enum.
    """

    ACCESS_CONTROL_REQUEST_METHOD = 'Access-Control-Request-Method'
    ACCESS_CONTROL_REQUEST_HEADERS = 'Access-Control-Request-Headers'


class CORSResponseHeaderEnum(CoreEnum):
    """
    cors response header enum.
    """

    ACCESS_CONTROL_ALLOW_ORIGIN = 'Access-Control-Allow-Origin'
    ACCESS_CONTROL_ALLOW_CREDENTIALS = 'Access-Control-Allow-Credentials'
    ACCESS_CONTROL_EXPOSE_HEADERS = 'Access-Control-Expose-Headers'
    ACCESS_CONTROL_MAX_AGE = 'Access-Control-Max-Age'
    ACCESS_CONTROL_ALLOW_METHODS = 'Access-Control-Allow-Methods'
    ACCESS_CONTROL_ALLOW_HEADERS = 'Access-Control-Allow-Headers'



# -*- coding: utf-8 -*-
"""
security exceptions module.
"""

from pyrin.core.exceptions import CoreBusinessException, CoreException
from pyrin.security.authentication.exceptions import AuthenticationFailedError


class SecurityManagerException(CoreException):
    """
    security manager exception.
    """
    pass


class SecurityManagerBusinessException(CoreBusinessException,
                                       SecurityManagerException):
    """
    security manager business exception.
    """
    pass


class InvalidUsernameOrPasswordError(AuthenticationFailedError,
                                     SecurityManagerBusinessException):
    """
    invalid username or password error.
    """
    pass


class UserNotFoundError(AuthenticationFailedError,
                        SecurityManagerBusinessException):
    """
    user not found error.
    """
    pass



from datetime import datetime

from demo.accounting.models import AccountEntity
from pyrin.api.schema.structs import ResultSchema
from pyrin.core.globals import SECURE_TRUE

i = []

class AccountSchema(ResultSchema):

    default_columns = ['owner']
    default_rename = dict(owner='Owner')

    def get_computed_row_columns(self, row, **options):
        """
        gets a dict containing all computed columns to be added to the result.

        this method is intended to be overridden in subclasses.

        :param ROW_RESULT row: the actual row result to be processed.

        :rtype: dict
        """

        if len(i) < 4:
            i.append(1)
            return dict(sum=1000,
                        date=datetime.now(),
                        entity=AccountEntity(id='my', owner='alireza', populate_all=SECURE_TRUE),
                        row=row,
                        formatted_id='ID: {id}'.format(id=row.id))

        return None

    def get_computed_entity_columns(self, entity, **options):
        """
        gets a dict containing all computed columns to be added to the result.

        this method is intended to be overridden in subclasses.

        :param BaseEntity entity: the actual entity to be processed.

        :rtype: dict
        """

        if len(i) < 4:
            i.append(1)
            return dict(sum=1000,
                        date=datetime.now(),
                        entity=AccountEntity(id='my', owner='alireza', populate_all=SECURE_TRUE),
                        formatted_id='ID: {id}'.format(id=entity.id))

        return None



# -*- coding: utf-8 -*-
"""
utils unique_id module.
"""

import uuid


def generate(**options):
    """
    generates a unique id based on input params.

    :keyword callable generator: uuid generator function.
                                 if not provided, defaults to uuid4.

    :rtype: uuid.UUID
    """

    generator = options.get('generator', uuid.uuid4)
    return generator()


def generate_uuid4():
    """
    generates a unique id using uuid4.

    :rtype: uuid.UUID
    """

    return generate(generator=uuid.uuid4)


def generate_uuid1():
    """
    generates a unique id using uuid1.

    :rtype: uuid.UUID
    """

    return generate(generator=uuid.uuid1)



# -*- coding: utf-8 -*-
"""
utils database module.
"""

from pyrin.utils.exceptions import InvalidSchemaNameError, InvalidTableNameError, \
    InvalidColumnNameError


def get_schema_name(schema_name, database_name=None):
    """
    gets fully qualified schema name for given database and schema.

    if database name is not provided, it returns the schema name only.

    note that for database backends which does not support cross
    database references, you should leave `database_name=None`.

    :param str schema_name: schema name.

    :param str database_name: database name.
                              if not provided, database name
                              will be excluded from result string.

    :raises InvalidSchemaNameError: invalid schema name error.

    :returns: fully qualified schema name in the form of:
              `database_name.schema_name` or `schema_name`

    :rtype: str
    """

    if database_name is None:
        database_name = ''

    if schema_name is None:
        schema_name = ''

    database_name = database_name.strip()
    schema_name = schema_name.strip()

    if schema_name == '':
        raise InvalidSchemaNameError('Schema name could not be blank.')

    if database_name == '':
        return schema_name

    return '{database}.{schema}'.format(database=database_name, schema=schema_name)


def get_column_name(schema_name, table_name, column_name, database_name=None):
    """
    gets fully qualified column name for given database and schema and table and column.

    note that for database backends which does not support cross
    database references, you should leave `database_name=None`.

    :param str schema_name: schema name.
    :param str table_name: table name.
    :param str column_name: column name.

    :param str database_name: database name.
                              if not provided, database name
                              will be excluded from result string.

    :raises InvalidTableNameError: invalid table name error.
    :raises InvalidColumnNameError: invalid column name error.

    :returns: fully qualified column name in the form of:
              `database_name.schema_name.table_name.column_name` or
              `schema_name.table_name.column_name`

    :rtype: str
    """

    qualified_schema = get_schema_name(schema_name, database_name)

    if table_name is None:
        table_name = ''

    if column_name is None:
        column_name = ''

    table_name = table_name.strip()
    column_name = column_name.strip()

    if table_name == '':
        raise InvalidTableNameError('Table name could not be blank.')

    if column_name == '':
        raise InvalidColumnNameError('Column name could not be blank.')

    return '{schema}.{table}.{column}'.format(schema=qualified_schema,
                                              table=table_name,
                                              column=column_name)



# -*- coding: utf-8 -*-
"""
authorization exceptions module.
"""

from pyrin.core.enumerations import ClientErrorResponseCodeEnum
from pyrin.core.exceptions import CoreBusinessException, CoreException
from pyrin.security.authentication.exceptions import AuthenticationFailedError


class AuthorizationManagerException(CoreException):
    """
    authorization manager exception.
    """
    pass


class AuthorizationManagerBusinessException(CoreBusinessException,
                                            AuthorizationManagerException):
    """
    authorization manager business exception.
    """
    pass


class AuthorizationFailedError(AuthorizationManagerBusinessException):
    """
    authorization failed error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of AuthorizationFailedError.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.FORBIDDEN


class UserNotAuthenticatedError(AuthenticationFailedError,
                                AuthorizationFailedError):
    """
    user not authenticated error.
    """
    pass


class UserIsNotActiveError(AuthorizationFailedError):
    """
    user is not active error.
    """
    pass



# -*- coding: utf-8 -*-
"""
security utils services module.
"""

from pyrin.application.services import get_component
from pyrin.security.utils import SecurityUtilsPackage


def generate_rsa_key(length=None, **options):
    """
    generates a pair of public/private rsa keys.

    :param int length: key length in bits.
                       if not provided, `rsa_default_key_length`
                       config will be used.

    :returns: tuple[str public_key, str private_key]
    :rtype: tuple[str, str]
    """

    return get_component(SecurityUtilsPackage.COMPONENT_NAME).generate_rsa_key(length,
                                                                               **options)


def load_rsa_key(public_pem, private_pem, **options):
    """
    loads public/private rsa key objects from provided value.

    :param str public_pem: public key content to load from.
    :param str private_pem: private key content to load from.

    :returns: tuple[object public_key, object private_key]
    :rtype: tuple[object, object]
    """

    return get_component(SecurityUtilsPackage.COMPONENT_NAME).load_rsa_key(public_pem,
                                                                           private_pem,
                                                                           **options)


def get_bytes(length=None, **options):
    """
    gets a secure random bytes with given length.

    the result value should not be decoded to string, because
    it's not a safe-string and may cause an error.
    if you want string representation, use `get_hex` or `get_url_safe` methods.

    :param int length: length of random bytes to be get.
                       if not provided, `default_secure_random_size`
                       config will be used.

    :rtype: bytes
    """

    return get_component(SecurityUtilsPackage.COMPONENT_NAME).get_bytes(length, **options)


def get_hex(length=None, **options):
    """
    gets a secure random hex string with given length.

    :param int length: length of random string to be get in bytes.
                       if not provided, `default_secure_random_size`
                       config will be used.

    :rtype: str
    """

    return get_component(SecurityUtilsPackage.COMPONENT_NAME).get_hex(length, **options)


def get_url_safe(length=None, **options):
    """
    gets a secure random url-safe string with given length.

    :param int length: length of random string to be get in bytes.
                       if not provided, `default_secure_random_size`
                       config will be used.

    :rtype: str
    """

    return get_component(SecurityUtilsPackage.COMPONENT_NAME).get_url_safe(length, **options)



# -*- coding: utf-8 -*-
"""
database paging paginator module.
"""

from copy import deepcopy
from abc import abstractmethod
from collections import OrderedDict

from flask import url_for

import pyrin.configuration.services as config_services
import pyrin.database.paging.services as paging_services
import pyrin.security.session.services as session_services

from pyrin.core.structs import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.database.paging.exceptions import PageSizeLimitError


class PaginatorBase(CoreObject):
    """
    paginator base class.
    """

    @abstractmethod
    def next(self):
        """
        gets the next page number.

        returns None if there is no next page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def previous(self):
        """
        gets the previous page number.

        returns None if there is no previous page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def inject_paging_keys(self, values, **options):
        """
        injects paging keys into given values from given inputs.

        :param dict values: dict values to inject paging keys into it.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def paginate(self, items, **options):
        """
        paginates the given items.

        it returns a tuple of two values, first value is a list of items
        to be returned to client, and second value is a dict of metadata
        to be injected into client response.

        :param list items: items to be paginated.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: tuple[list items, dict metadata]
        :rtype: tuple[list, dict]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has_next(self, count, *args, **options):
        """
        gets a value indicating that there is a next page available.

        it returns a tuple of two items. first item is a boolean indicating
        that there is a next page and the second item is the number of excess
        items that must be removed from end of items.

        :param int count: count of current items.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: tuple[bool has_next, int excess]
        :rtype: tuple[bool, int]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has_previous(self, count, *args, **options):
        """
        gets a value indicating that there is a previous page available.

        it returns a tuple of two items. first item is a boolean indicating
        that there is a previous page and the second item is the number of
        excess items that must be removed from beginning of items.

        :param int count: count of current items.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: tuple[bool has_previous, int excess]
        :rtype: tuple[bool, int]
        """

        raise CoreNotImplementedError()

    def copy(self):
        """
        returns a deep copy of this instance

        :rtype: PaginatorBase
        """

        return deepcopy(self)

    @property
    @abstractmethod
    def current_page(self):
        """
        gets current page number.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def current_page_size(self):
        """
        gets current page size.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()


class SimplePaginator(PaginatorBase):
    """
    simple paginator class.

    page numbers start from 1.
    it does not emit any extra queries to database to fetch count or like that.
    the only limitation is that it could not detect previous page from the
    `last_page + 1` page.
    """

    def __init__(self, endpoint, **options):
        """
        initializes an instance of SimplePaginator.

        :param str endpoint: endpoint of route.

        :keyword int page_size: default page size.
                                if not provided, it will be get from
                                `default_page_size` of `database` config store.

        :keyword int max_page_size: max allowed page size.
                                    if not provided, it will be get from
                                    `max_page_size` of `database` config store.

        :raises PageSizeLimitError: page size limit error.
        """

        super().__init__()

        global_max_page_size = config_services.get('database', 'paging', 'max_page_size')
        max_page_size = options.get('max_page_size')
        if max_page_size is None or max_page_size < 1:
            max_page_size = global_max_page_size

        if max_page_size > global_max_page_size:
            raise PageSizeLimitError('Max page size [{max}] is bigger than global max page '
                                     'size which is [{global_max}] on endpoint [{endpoint}].'
                                     .format(max=max_page_size,
                                             global_max=global_max_page_size,
                                             endpoint=endpoint))

        page_size = options.get('page_size')
        default_page_size = config_services.get('database', 'paging', 'default_page_size')
        if page_size is None or page_size < 1:
            page_size = min(default_page_size, max_page_size)

        if page_size > max_page_size:
            raise PageSizeLimitError('Page size [{page_size}] is bigger than max page size '
                                     'which is [{max}] on endpoint [{endpoint}].'
                                     .format(page_size=page_size,
                                             max=max_page_size,
                                             endpoint=endpoint))

        self._page_size = page_size
        self._max_page_size = max_page_size
        self._endpoint = endpoint
        self._limit = None
        self._offset = None
        self._current_page = None
        self._current_page_size = None
        self._has_next = False
        self._has_previous = False

    def _url_for(self, page, page_size):
        """
        gets the url for given page number.

        :param int page: page number to generate its url.
        :param int page_size: page size.

        :rtype: str
        """

        request = session_services.get_current_request()
        options = OrderedDict()
        options.update(paging_services.generate_paging_params(page, page_size))
        options.update(request.get_all_query_strings())
        return url_for(self._endpoint, **options)

    def has_next(self, count, **options):
        """
        gets a value indicating that there is a next page available.

        it returns a tuple of two items. first item is a boolean indicating
        that there is a next page and the second item is the number of excess
        items that must be removed from end of items.

        :param int count: count of current items.

        :returns: tuple[bool has_next, int excess]
        :rtype: tuple[bool, int]
        """

        # the original limit is always 2 less than the current limit.
        excess = count - (self._limit - 2)
        if excess <= 0:
            self._has_next = False
            return self._has_next, 0

        if self._current_page == 1:
            self._has_next = excess > 0
            return self._has_next, excess
        else:
            self._has_next = excess > 1
            return self._has_next, excess - 1

    def has_previous(self, count, **options):
        """
        gets a value indicating that there is a previous page available.

        it returns a tuple of two items. first item is a boolean indicating
        that there is a previous page and the second item is the number of
        excess items that must be removed from beginning of items.

        :param int count: count of current items.

        :returns: tuple[bool has_previous, int excess]
        :rtype: tuple[bool, int]
        """

        # at any page, if there is a count > 0, it means that there is a previous
        # page available. because the first item is from the previous page.
        if count <= 0 or self._current_page == 1:
            self._has_previous = False
            return self._has_previous, 0

        self._has_previous = True
        return self._has_previous, 1

    def next(self):
        """
        gets the next page number.

        returns None if there is no next page.

        :rtype: int
        """

        if self._has_next is True:
            return self._url_for(self._current_page + 1, self._current_page_size)

        return None

    def previous(self):
        """
        gets the previous page number.

        returns None if there is no previous page.

        :rtype: int
        """

        if self._has_previous is True:
            return self._url_for(self._current_page - 1, self._current_page_size)

        return None

    def inject_paging_keys(self, values, **options):
        """
        injects paging keys into given values from given inputs.

        :param dict values: dict values to inject paging keys into it.

        :keyword int page: page number.
        :keyword int page_size: page size.
        """

        page, page_size = paging_services.get_paging_params(**options)

        if page is None or not isinstance(page, int) or page < 1:
            page = 1

        if page_size is None or not isinstance(page_size, int) or page_size < 1:
            page_size = self._page_size
        elif page_size > self._max_page_size:
            page_size = self._max_page_size

        # we increase limit by 2 to be able to detect if there is a next and previous page.
        # the extra items will not be returned to client.
        self._limit = page_size + 2
        offset = page - 1
        extra_offset = offset * page_size
        if extra_offset > 0:
            # we decrease offset by 1 to be able to detect if there is a previous page.
            # the extra item will not be returned to client.
            extra_offset = extra_offset - 1

        self._offset = extra_offset
        self._current_page = page
        self._current_page_size = page_size
        paging_services.inject_paging_keys(self._limit, self._offset, values)

    def paginate(self, items, **options):
        """
        paginates the given items.

        it returns a tuple of two values, first value is a list of items
        to be returned to client, and second value is a dict of metadata
        to be injected into client response.

        :param list items: items to be paginated.

        :returns: tuple[list items, dict metadata]
        :rtype: tuple[list, dict]
        """

        metadata = OrderedDict()
        count = len(items)
        result = items

        has_next, excess_end = self.has_next(count)
        has_previous, excess_first = self.has_previous(count)
        if has_next is True:
            result = result[:-excess_end]
            count = count - excess_end

        if has_previous is True:
            result = result[excess_first:]
            count = count - excess_first

        next_url = self.next()
        previous_url = self.previous()
        metadata.update(count=count, next=next_url, previous=previous_url)
        return result, metadata

    @property
    def current_page(self):
        """
        gets current page number.

        :rtype: int
        """

        return self._current_page

    @property
    def current_page_size(self):
        """
        gets current page size.

        :rtype: int
        """

        return self._current_page_size



# -*- coding: utf-8 -*-
"""
security manager module.
"""

import pyrin.security.hashing.services as hashing_services
import pyrin.security.token.services as toke_services
import pyrin.globalization.datetime.services as datetime_services

from pyrin.core.globals import _
from pyrin.core.structs import DTO
from pyrin.database.services import get_current_store
from pyrin.security.authorization.exceptions import UserIsNotActiveError
from pyrin.security.manager import SecurityManager as BaseSecurityManager

from demo.security.exceptions import UserNotFoundError, InvalidUsernameOrPasswordError
from demo.security.models import UserEntity


class SecurityManager(BaseSecurityManager):
    """
    security manager class.
    this class is intended to provide some services needed in pyrin application.
    the top level application must extend this class considering business requirements.
    """

    def has_permission(self, user, permissions, **options):
        """
        gets a value indicating that given user has the specified permissions.

        :param user: user identity to check its permissions.
        :param list[PermissionBase] permissions: permissions to check for user.

        :rtype: bool
        """

        return False

    def login(self, username, password, **options):
        """
        logs in the provided user and gets a valid token.

        :param str username: username.
        :param str password: password.

        :raises InvalidUserInfoError: invalid user info error.
        :raises UserNotFoundError: user not found error.
        :raises UserIsNotActiveError: user is not active error.

        :returns: a valid token.
        :rtype: str
        """

        if username is None or password is None:
            raise InvalidUsernameOrPasswordError(_('Username or password is invalid.'))

        store = get_current_store()
        user = store.query(UserEntity).filter(UserEntity.username == username).one_or_none()

        fail_message = _('User not found, make sure that '
                         'username and password is correct.')
        if user is None:
            raise UserNotFoundError(fail_message)

        if hashing_services.is_match(password, user.password_hash) is not True:
            raise UserNotFoundError(fail_message)

        if user.is_active is False:
            raise UserIsNotActiveError(_('User [{user_id}] is not active.'
                                         .format(user_id=user.id)))

        user.last_login_date = datetime_services.now()
        store.add(user)
        payload = DTO(user_id=user.id)

        return DTO(access_token=toke_services.generate_access_token(payload, is_fresh=True),
                   refresh_token=toke_services.generate_refresh_token(payload))



# -*- coding: utf-8 -*-
"""
response component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.processor.response import ResponsePackage
from pyrin.processor.response.manager import ResponseManager


@component(ResponsePackage.COMPONENT_NAME)
class ResponseComponent(Component, ResponseManager):
    """
    response component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
security services module.
"""

from pyrin.application.services import get_component

from demo.security import SecurityPackage


def login(username, password, **options):
    """
    logs in the provided user and gets a valid token.

    :param str username: username.
    :param str password: password.

    :raises InvalidUserInfoError: invalid user info error.
    :raises UserNotFoundError: user not found error.
    :raises UserIsNotActiveError: user is not active error.

    :returns: a valid token.
    :rtype: str
    """

    return get_component(SecurityPackage.COMPONENT_NAME).login(username, password, **options)


import pyrin.security.services as security_services

security_services.has_permission(1, [2])
security_services.has_permission(1, [2])
security_services.has_permission(1, [2])
security_services.has_permission(1, [2])



# -*- coding: utf-8 -*-
"""
caching interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.caching.exceptions import KeyIsNotPresentInCacheError
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject, MultiSingletonMeta


class CacheSingletonMeta(MultiSingletonMeta):
    """
    cache singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractCache(CoreObject, metaclass=CacheSingletonMeta):
    """
    abstract cache class.

    all application caches must be subclassed from this.
    """

    def __setitem__(self, key, value):
        """
        sets the given key with given value into this cache.

        :param object key: key to set value with it.
        :param object value: value to be cached.
        """

        return self.set(key, value)

    def __getitem__(self, key):
        """
        gets the value of given key.

        it raises an error if key does not exist.

        :param object key: key to get its value.

        :raises KeyIsNotPresentInCacheError: key is not present in cache error.

        :returns: object
        """

        result = self.get(key, default=KeyIsNotPresentInCacheError)
        if result is KeyIsNotPresentInCacheError:
            raise KeyIsNotPresentInCacheError('Key [{key}] is not present in '
                                              'the cache.'.format(key=key))

        return result

    def __contains__(self, key):
        """
        gets a value indicating that given key exists in cache.

        :param object key: key to be checked for existence.

        :rtype: bool
        """

        return self.contains(key)

    @abstractmethod
    def set(self, key, value, *args, **options):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get(self, key, default=None, **options):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def try_set(self, value, *keys, **options):
        """
        sets a new value into cached items.

        this method will generate cache key from given type and function.
        if the provided values are not hashable, this method won't raise
        an error and logs it silently.

        :param object value: value to be cached.
        :param object keys: value to be used for key generation.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def try_get(self, *keys, default=None, **options):
        """
        gets the value from cache.

        this method will generate cache key from given type and function.
        if key does not exist, it returns None or the specified default value.
        if the provided values are not hashable, this method won't raise
        an error and logs it silently.

        :param object keys: value to be used for key generation.
        :param object default: value to be returned if key is not present.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def contains(self, key):
        """
        gets a value indicating that given key is existed in the cached items.

        :param object key: key to be checked for existence.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def pop(self, key, default=None):
        """
        pops the given key from cache and returns its value.

        if key does not exist, it returns None or the specified default value.

        :param object key: key to get its value.
        :param object default: value to be returned if key is not present.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def remove(self, key, **options):
        """
        removes the given key from cache.

        it does nothing if the key is not present in the cache.

        :param object key: key to be removed.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def clear(self):
        """
        clears all items from cache.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def generate_key(self, *keys, **options):
        """
        generates a cache key from given inputs.

        :param object keys: value to be used for key generation.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: hash of generated key.
        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def last_cleared_time(self):
        """
        gets the last time in which this cache has been cleared.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: datetime.datetime
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def stats(self):
        """
        get the statistic info about cached items.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def persistent(self):
        """
        gets a value indicating that cached items must be persisted to database on shutdown.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()


class AbstractLocalCache(AbstractCache):
    """
    abstract local cache class.

    all application local caches must be subclassed from this.

    this type of caches does not consider method inputs, current user and
    component key in key generation. it only considers the class type
    of function and function name itself. this is useful for caching items
    that never change after application startup and are independent
    from different scoped or global variables.

    it also does not support expire time and size limit for cached values.
    its values are permanent unless manually removed if required.

    it also keeps the real value in the cache, not a deep copy of it to gain
    performance.

    it also does not provide statistic info about hit or missed
    caches, to gain performance.
    """

    def __len__(self):
        """
        gets the count of items of this cache.

        :rtype: int
        """

        return self.count

    @abstractmethod
    def items(self):
        """
        gets an iterable of all keys and their values in the cache.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterable[tuple[object, object]]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def keys(self):
        """
        gets all keys of current cache.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterable[object]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def values(self):
        """
        gets all values of current cache.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: iterable[object]
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def count(self):
        """
        gets the count of items of this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()


class AbstractExtendedLocalCache(AbstractLocalCache):
    """
    abstract extended local cache class.

    all application extended local caches must be subclassed from this.

    this type of caches are same as `AbstractLocalCache` type but it also
    considers method inputs, current user and component key.
    """

    @property
    @abstractmethod
    def consider_user(self):
        """
        gets the consider user value for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()


class AbstractComplexLocalCache(AbstractExtendedLocalCache):
    """
    abstract complex local cache class.

    all application complex local caches must be subclassed from this.

    this type of caches will also consider method inputs, current user
    and component key in key generation. this is useful for caching items that
    change during application runtime based on different inputs and variables.

    it also supports expire time and size limit for cached items.
    it also keeps a deep copy of the value in the cache.
    it also provides statistic info about hit or missed caches.
    it also supports persistent mode to save cached values into
    database on application shutdown and load them back on next startup.
    """

    @abstractmethod
    def persist(self, version, **options):
        """
        saves cached items of this cache into database.

        :param str version: version to be saved with cached items in database.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def load(self, version, **options):
        """
        loads cached items of this cache from database.

        :param str version: version of cached items to be loaded from database.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def is_full(self):
        """
        gets a value indicating the cache is full.

        it returns a tuple, first item is a boolean indicating the fullness of cache.
        the second item is the number of excess items in the cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: tuple[bool, int]
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def expire(self):
        """
        gets default expire time value for this cache's items in milliseconds.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def limit(self):
        """
        gets the count limit of this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def refreshable(self):
        """
        gets a value indicating that cached item's expire time will be extended on each hit.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def hit_count(self):
        """
        gets the hit count for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def miss_count(self):
        """
        gets the miss count for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def use_lifo(self):
        """
        gets the use lifo for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def clear_count(self):
        """
        gets the clear count for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def chunk_size(self):
        """
        gets the chunk size for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()


class AbstractRemoteCache(AbstractCache):
    """
    abstract remote cache class.

    all application remote caches must be subclassed from this.
    """

    @property
    @abstractmethod
    def expire(self):
        """
        gets default expire time value for this cache's items in milliseconds.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def hit_count(self):
        """
        gets the hit count for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def miss_count(self):
        """
        gets the miss count for this cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
permission package.
"""

from pyrin.security.permission import PermissionPackage as BasePermissionPackage


class PermissionPackage(BasePermissionPackage):
    """
    permission package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
permission manager module.
"""

from pyrin.database.services import get_current_store
from pyrin.security.permission.manager import PermissionManager as BasePermissionManager

from demo.security.permission.models import PermissionEntity


class PermissionManager(BasePermissionManager):
    """
    permission manager class.
    """

    def _exists(self, permission_id):
        """
        gets a value indicating that given permission exists in database.

        :param int permission_id: permission id.

        :rtype: bool
        """

        store = get_current_store()
        return store.query(PermissionEntity.id).filter(PermissionEntity.id ==
                                                       permission_id).existed()



# -*- coding: utf-8 -*-
"""
audit services module.
"""

from pyrin.audit import AuditPackage
from pyrin.application.services import get_component


def register_hook(instance):
    """
    registers the given instance into audit hooks.

    :param AuditHookBase instance: audit hook instance to be registered.

    :raises InvalidAuditHookTypeError: invalid audit hook type error.
    """

    return get_component(AuditPackage.COMPONENT_NAME).register_hook(instance)


def inspect(**options):
    """
    inspects all registered packages and gets inspection data.

    it includes all available info, but it could be customized using different options.
    other custom implementations could also be excluded from result if the related name
    of their hook is passed with False value. for example `database=False`.

    it returns a tuple of two values. the first value is a dict containing the
    inspection data, and the second value is the related status code for the response.
    in case of any inspection has been failed, the status code will be internal
    server error 500.

    :keyword bool application: specifies that application info must be included.
    :keyword bool packages: specifies that loaded packages info must be included.
    :keyword bool framework: specifies that framework info must be included.
    :keyword bool python: specifies that python info must be included.
    :keyword bool os: specifies that operating system info must be included.
    :keyword bool hardware: specifies that hardware info must be included.

    :keyword bool traceback: specifies that on failure report, it must include
                             the traceback of errors.
                             defaults to True if not provided.

    :keyword bool raise_error: specifies that it must raise error
                               if any of registered audits failed
                               instead of returning a failure response.
                               defaults to False if not provided.

    :raises AuditFailedError: audit failed error.

    :returns: tuple[dict(dict application: application info,
                         dict packages: loaded packages info,
                         dict framework: framework info,
                         dict python: python info,
                         dict platform: platform info),
                    int status_code]
    :rtype: tuple[dict, int]
    """

    return get_component(AuditPackage.COMPONENT_NAME).inspect(**options)


def startup_inspect():
    """
    inspects all registered packages on application startup.

    it raises an error if anything goes wrong.

    :raises AuditFailedError: audit failed error.
    """

    return get_component(AuditPackage.COMPONENT_NAME).startup_inspect()


def get_application_info(**options):
    """
    gets the info of current application.

    :returns: dict(str name: application name,
                   datetime datetime: application current datetime,
                   str version: application version)
    :rtype: dict
    """

    return get_component(AuditPackage.COMPONENT_NAME).get_application_info(**options)


def get_framework_info(**options):
    """
    gets the info of current framework.

    :returns: dict(str name: framework name,
                   str version: framework version)
    :rtype: dict
    """

    return get_component(AuditPackage.COMPONENT_NAME).get_framework_info(**options)


def get_python_info(**options):
    """
    gets the current python version info which application is running on it.

    :returns: dict(str version: python version,
                   str implementation: python implementation)
    :rtype: dict
    """

    return get_component(AuditPackage.COMPONENT_NAME).get_python_info(**options)


def get_operating_system_info(**options):
    """
    gets the current operating system info.

    :returns: dict(str name: os name,
                   str release: os release,
                   str version: os version)
    :rtype: dict
    """

    return get_component(AuditPackage.COMPONENT_NAME).get_operating_system_info(**options)


def get_hardware_info(**options):
    """
    gets the current machine's hardware info.

    :returns: dict(str processor: processor name,
                   str machine: machine name)
    :rtype: dict
    """

    return get_component(AuditPackage.COMPONENT_NAME).get_hardware_info(**options)


def get_platform_info(**options):
    """
    gets the platform info of current machine.

    it includes all available info, but it could be customized using different options.

    :keyword bool os: specifies that operating system info must be included.
    :keyword bool hardware: specifies that hardware info must be included.

    :returns: dict(dict os: os info,
                   dict hardware: hardware info)
    :rtype: dict
    """

    return get_component(AuditPackage.COMPONENT_NAME).get_platform_info(**options)


def get_audit_configurations():
    """
    gets the audit api configurations.

    :returns: dict(bool enabled: enable audit api,
                   bool authenticated: audit api access type,
                   int request_limit: audit api request count limit,
                   str url: audit api exposed url)
    :rtype: dict
    """

    return get_component(AuditPackage.COMPONENT_NAME).get_audit_configurations()



# -*- coding: utf-8 -*-
"""
model component module.
"""

from pyrin.application.decorators import component
from pyrin.database.model import ModelPackage
from pyrin.database.model.manager import ModelManager
from pyrin.application.structs import Component


@component(ModelPackage.COMPONENT_NAME)
class ModelComponent(Component, ModelManager):
    """
    model component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
validator handlers number module.
"""

from pyrin.core.globals import _
from pyrin.validator.handlers.base import ValidatorBase
from pyrin.validator.handlers.exceptions import ValueIsNotNumberError, ValueIsNotIntegerError, \
    ValueIsNotFloatError


class NumberValidator(ValidatorBase):
    """
    number validator class.
    """

    invalid_type_error = ValueIsNotNumberError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be a number.')

    def __init__(self, domain, name, **options):
        """
        initializes an instance of NumberValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        options.update(accepted_type=(int, float))
        super().__init__(domain, name, **options)


class IntegerValidator(ValidatorBase):
    """
    integer validator class.
    """

    invalid_type_error = ValueIsNotIntegerError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be an integer.')

    def __init__(self, domain, name, **options):
        """
        initializes an instance of IntegerValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        options.update(accepted_type=int)
        super().__init__(domain, name, **options)


class FloatValidator(ValidatorBase):
    """
    float validator class.
    """

    invalid_type_error = ValueIsNotFloatError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be a float number.')

    def __init__(self, domain, name, **options):
        """
        initializes an instance of FloatValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        options.update(accepted_type=float)
        super().__init__(domain, name, **options)



# -*- coding: utf-8 -*-
"""
utils string module.
"""


def upper(value):
    """
    returns the uppercase copy of input string.

    :param str value: string to make uppercase.

    :rtype: str
    """

    return value.upper()


def lower(value):
    """
    returns the lowercase copy of input string.

    :param str value: string to make lowercase.

    :rtype: str
    """

    return value.lower()


def remove_line_break_escapes(value):
    """
    removes line break escapes from given value.
    it replaces `\\n` with `\n` to enable line breaks.

    :param str value: value to remove line break escapes from it.

    :rtype: str
    """

    return value.replace('\\n', '\n')


def remove_duplicate_space(value):
    """
    removes all duplicate spaces and keeps just a single space in given value.

    :param str value: value to remove duplicate spaces from it.

    :rtype: str
    """

    if value is None:
        return None

    while '  ' in value:
        value = value.replace('  ', ' ')

    return value


def interpolate(message, data):
    """
    interpolates the given message with given data.

    if data is not a dict, no interpolation will be done.

    :param str message: message to be interpolated.
    :param dict data: data to be used for interpolation.

    :rtype: str
    """

    if not isinstance(data, dict):
        return message

    return message.format(**data)


def sort_by_length(items, reverse=False):
    """
    sorts the given list of strings by length of items.

    :param list[str] items: list of strings to be sorted.

    :param bool reverse: sort by descending length.
                         defaults to False if not provided.

    :rtype: list[str]
    """

    return sorted(items, key=len, reverse=reverse)


def quote(value, sign="'"):
    """
    quotes the given string value.

    :param str value: value to be quoted.
    :param str sign: quotation sign to be used.
                     defaults to single quotation if not provided.

    :rtype: str
    """

    return "{sign}{value}{sign}".format(sign=sign, value=str(value))


def union(first, second, **options):
    """
    gets the list of strings which are in both lists.

    note that it returns the value from the first list.
    by default comparison is case-insensitive.
    so if `A` is in the first list and `a` is in the second list, the
    result list will contain `A` from the first list.

    :param list[str] | set[str] | tuple[str] first: first list of strings.
    :param list[str] | set[str] | tuple[str] second: second list of strings.

    :keyword function converter: a callable to be used for case conversion of
                                 strings. it must take a single input string and
                                 return the converted value. if you want case-sensitive
                                 comparison, you could pass `str` function as converter.
                                 defaults to `lower` which results in case-insensitive
                                 comparison.

    :keyword type collection: collection type to be used for result.
                              defaults to `list` if not provided.

    :rtype: list[str] | set[str] | tuple[str]
    """

    collection = options.get('collection', list)
    converter = options.get('converter', lower)
    result = []
    second = [converter(item) for item in second]
    for item in first:
        if converter(item) in second:
            result.append(item)

    return collection(result)


def is_match(item, values, **options):
    """
    gets a value indicating that given item exists in the given list of values.

    by default comparison is case-insensitive.
    so if item is `A` and `a` is in the values list, it returns True.

    :param str item: value to be checked for matching.
    :param list[str] | set[str] | tuple[str] values: list of string values.

    :keyword function converter: a callable to be used for case conversion of
                                 strings. it must take a single input string and
                                 return the converted value. if you want case-sensitive
                                 comparison, you could pass `str` function as converter.
                                 defaults to `lower` which results in case-insensitive
                                 comparison.

    :rtype: bool
    """

    result = union([item], values, **options)
    return len(result) > 0



# -*- coding: utf-8 -*-
"""
sentry integrations package.
"""

from pyrin.packaging.base import Package


class SentryIntegrationsPackage(Package):
    """
    sentry integrations package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
celery cli services module.
"""

from pyrin.application.services import get_component
from pyrin.task_queues.celery.cli import CeleryCLIPackage


def register_cli_handler(instance, **options):
    """
    registers a new celery cli handler or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a cli handler which is already registered.

    :param CeleryCLIHandlerBase instance: celery cli handler to be registered.
                                          it must be an instance of
                                          CeleryCLIHandlerBase.

    :keyword bool replace: specifies that if there is another registered
                           cli handler with the same name, replace it
                           with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidCLIHandlerTypeError: invalid cli handler type error.
    :raises DuplicatedCLIHandlerError: duplicated cli handler error.
    """

    get_component(CeleryCLIPackage.COMPONENT_NAME).register_cli_handler(instance, **options)


def execute(handler_name, **options):
    """
    executes the handler with the given name with given inputs.

    :param str handler_name: handler name to be executed.

    :raises CLIHandlerNotFoundError: cli handler not found error.
    """

    return get_component(CeleryCLIPackage.COMPONENT_NAME).execute(handler_name, **options)


def execute_on_celery(name, arguments=None):
    """
    executes the given celery command.

    :param str name: celery command name to be executed.
    :param list arguments: all command line arguments.

    :returns: result of celery command execution.
    """

    return get_component(CeleryCLIPackage.COMPONENT_NAME).execute_on_celery(name, arguments)


def get_package_class():
    """
    gets the package class of celery cli manager.

    :raises PackageClassIsNotSetError: package class is not set error.

    :returns: type[CeleryCLIPackage]
    """

    return get_component(CeleryCLIPackage.COMPONENT_NAME).get_package_class()



# -*- coding: utf-8 -*-
"""
application hooks module.
"""

import pyrin.application.services as application_services

from pyrin.core.structs import Hook
from pyrin.configuration import ConfigurationPackage
from pyrin.packaging.hooks import PackagingHookBase


class ApplicationHookBase(Hook):
    """
    application hook base class.

    all packages that need to be hooked into application business must
    implement this class and register it in application hooks.
    """

    def after_application_loaded(self):
        """
        this method will be called after application has been loaded.
        """
        pass

    def application_initialized(self):
        """
        this method will be get called after application has been fully initialized.
        """
        pass

    def prepare_runtime_data(self):
        """
        this method will be get called after application has been fully initialized.

        any changes of this method to database, will be committed automatically by
        application. so you should not commit anything in the hook. if you do commit
        manually, unexpected behaviors may occur.

        note that this method will not get called when application starts in scripting mode.
        """
        pass

    def before_application_run(self):
        """
        this method will be get called just before application gets running.

        note that this method will not get called when
        application starts in scripting mode.
        """
        pass

    def application_status_changed(self, old_status, new_status):
        """
        this method will be called whenever application status changes.

        :param str old_status: old application status.
        :param str new_status: new application status.

        :enum status:
            INITIALIZING = 'Initializing'
            LOADING = 'Loading'
            READY = 'Ready'
            RUNNING = 'Running'
            TERMINATED = 'Terminated'
        """
        pass

    def provide_response_headers(self, headers, endpoint,
                                 status_code, method, **options):
        """
        this method will be called whenever a response is going to be returned from server.

        subclasses could override this to provide custom or modified response headers.
        they must modify or extend headers in-place.

        :param dict | Headers headers: current response headers.

        :param str endpoint: the endpoint of the route that
                             handled the current request.
                             by default, it is the fully qualified
                             name of the view function.

        :param int status_code: response status code.
                                it could be None if not provided.

        :param str method: the http method of current request.

        :keyword str url: the url of the route that handled this request.

        :keyword user: the user of current request.
                       it could be None.
        """
        pass

    def finalize_transaction(self, response, **options):
        """
        this method will be called after a request has been fully processed.

        note that this method will be called after all `after_request_handlers`
        have been executed. even if any handler had error, this method call is
        guaranteed.
        subclasses must return the same input response, a modified response or a
        new response object. but if they don't, it will be converted to a response.
        note that it is recommended for subclasses not to raise any exceptions
        in this method. but if they do so, other handlers will still get executed.
        the main use case for this hook is to finalize any present transactions,
        such as database.

        :param CoreResponse response: response object.

        :rtype: CoreResponse
        """

        return response

    def validate_request(self, request, **options):
        """
        this method will be called whenever a request is received before any other operation.

        subclasses could use this to validate request and if required, raise an error
        to prevent request progression.
        if any subclass raises an error during validation, all hooks, request handlers
        and url processors will be ignored and error response will be returned to client
        immediately.
        note that the returned value of this method will be ignored.

        :param CoreRequest request: current request instance.
        """
        pass


class PackagingHook(PackagingHookBase):
    """
    packaging hook class.
    """

    def package_loaded(self, package_name, **options):
        """
        this method will be called after each application package has been loaded.

        :param str package_name: name of the loaded package.
        """

        # we should call this method as soon as configuration package is
        # loaded, to make sure application configs are loaded before any
        # other package needs them.
        if package_name == ConfigurationPackage.NAME:
            application_services.load_configs(**options)



# -*- coding: utf-8 -*-
"""
orm scoping registry interface module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject


class AbstractAtomicContainerBase(CoreObject):
    """
    abstract atomic container base class.
    """

    @abstractmethod
    def has(self):
        """
        gets a value indicating that an atomic object is present in the current scope.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def set(self, obj):
        """
        sets an atomic value for the current scope.

        :param object obj: object to be set in registry.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def clear(self, atomic=False):
        """
        clears all objects of current scope, if any.

        if `atomic=True` is provided, it only clears the
        current atomic object if available.

        :param bool atomic: specifies that it must only clear the current atomic
                            object of current scope. otherwise, it clears all atomic
                            objects of current scope. defaults to False if not provided.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get(self):
        """
        gets the current atomic object of current scope if available, otherwise returns None.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_all(self):
        """
        gets all atomic objects of current scope if available, otherwise returns an empty list.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: list[object]
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
serializer component module.
"""

from pyrin.converters.serializer import SerializerPackage
from pyrin.converters.serializer.manager import SerializerManager
from pyrin.application.decorators import component
from pyrin.application.structs import Component


@component(SerializerPackage.COMPONENT_NAME)
class SerializerComponent(Component, SerializerManager):
    """
    serializer component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching remote package.
"""

from pyrin.packaging.base import Package


class CachingRemotePackage(Package):
    """
    caching remote package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
audit enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class InspectionStatusEnum(CoreEnum):
    """
    inspection status enum.
    """

    OK = 'OK'
    FAILED = 'Failed'



# -*- coding: utf-8 -*-
"""
response services module.
"""

from pyrin.processor.response import ResponsePackage
from pyrin.application.services import get_component


def make_response(**options):
    """
    makes a response from given inputs.

    note that every other keyword argument that is present
    in given input, will be passed to generated response.

    :keyword int code: response code.
                       defaults to `DEFAULT_STATUS_CODE`
                       code, if not provided.

    :keyword dict | Headers headers: headers to add into response.

    :returns: tuple[dict | object, int, CoreHeaders]
    :rtype: tuple
    """

    return get_component(ResponsePackage.COMPONENT_NAME).make_response(**options)


def make_error_response(message, **options):
    """
    makes an error response from given inputs.

    :param str message: error message.

    :keyword int code: error code.
                       defaults to `INTERNAL_SERVER_ERROR`
                       code, if not provided.

    :keyword dict | Headers headers: headers to add into response.

    :returns: tuple[dict | object, int, CoreHeaders]
    :rtype: tuple
    """

    return get_component(ResponsePackage.COMPONENT_NAME).make_error_response(message,
                                                                             **options)


def make_exception_response(exception, **options):
    """
    makes an error response from given exception.

    if the exception does not have code, defaults
    to `INTERNAL_SERVER_ERROR` code.

    :param Exception exception: exception instance.

    :keyword int code: error code.
                       tries to get it from exception itself, if not
                       provided. defaults to `INTERNAL_SERVER_ERROR`
                       code if not available in exception.

    :keyword dict | Headers headers: headers to add into response.

    :returns: tuple[dict | object, int, CoreHeaders]
    :rtype: tuple
    """

    return get_component(ResponsePackage.COMPONENT_NAME).make_exception_response(exception,
                                                                                 **options)


def unpack_response(response, **options):
    """
    unpacks the response object into a tuple of three parts.

    in the form of `(body, status_code, headers)`. if any of these
    parts are not present in provided response, it returns None for
    that specific part.
    note that if a view function returns a `Response` object, the status
    code will be fetched from that object if available. but if a view function
    returns (Response response, int status_code) then the stand-alone status
    code will override the status code of `Response` object.

    :param tuple | object response: response object to be unpacked.

    :returns: tuple[object body, int status_code, CoreHeaders headers]
    :rtype: tuple[object, int, CoreHeaders]
    """

    return get_component(ResponsePackage.COMPONENT_NAME).unpack_response(response,
                                                                         **options)


def pack_response(body, status_code, headers, **options):
    """
    packs the response using given values.

    it returns a tuple if status code or headers are
    not None, otherwise it just returns the body.

    :param object | CoreResponse body: body of response.
    :param int status_code: status code of response.
    :param dict | Headers headers: all response headers.

    :returns: tuple[object body, int status_code, CoreHeaders headers] | object
    :rtype: tuple[object, int, CoreHeaders] | object
    """

    return get_component(ResponsePackage.COMPONENT_NAME).pack_response(body, status_code,
                                                                       headers, **options)


def get_body(response, **options):
    """
    gets the first part of a response tuple.

    it could be the body data or a response object itself.
    if response is not a tuple, the return value is the response itself.

    :param tuple | CoreResponse | object response: the response object
                                                   or instance or tuple.

    :rtype: CoreResponse | object
    """

    return get_component(ResponsePackage.COMPONENT_NAME).get_body(response, **options)



# -*- coding: utf-8 -*-
"""
caching remote handlers memcached module.
"""

import pickle

from pymemcache import Client

from pyrin.caching.decorators import cache
from pyrin.caching.exceptions import InvalidCacheLimitError
from pyrin.caching.globals import NO_LIMIT
from pyrin.caching.remote.handlers.base import RemoteCacheBase
from pyrin.caching.remote.handlers.exceptions import HostnameOrUnixSocketRequiredError, \
    BothHostnameAndUnixSocketProvidedError


@cache()
class Memcached(RemoteCacheBase):
    """
    memcached class.
    """

    cache_name = 'memcached'

    HIT_KEY = '__HIT_COUNT__'
    MISS_KEY = '__MISS_COUNT__'

    def __init__(self, *args, **options):
        """
        initializes an instance of Memcached.

        :keyword int limit: limit for memory usage by cached items in megabytes.
                            if not provided, it will be get from
                            `caching` config store. if you want to
                            remove size limit, you could pass
                            `caching.globals.NO_LIMIT` as input. note
                            that this will remove limit only by client,
                            so memcached server default limit will be intact.

        :keyword int expire: default expire time of cached items in seconds.
                             if not provided, it will be get from `caching` config
                             store.

        :keyword bool consider_user: specifies that current user must also be
                                     included in cache key. if not provided, will
                                     be get from `caching` config store.

        :raises CacheNameIsRequiredError: cache name is required error.
        :raises InvalidCacheExpireTimeError: invalid cache expire time error.
        :raises InvalidCacheLimitError: invalid cache limit error.
        """

        self._limit = NO_LIMIT

        super().__init__(*args, **options)

        self.client.add(self.HIT_KEY, 0, expire=0, noreply=True, flags=1)
        self.client.add(self.MISS_KEY, 0, expire=0, noreply=True, flags=1)

    def _create_client(self, *args, kwargs=None, **configs):
        """
        creates a client for connecting to memcached server.

        :param object args: all positional arguments passed to `__init__` method.
        :param dict kwargs: all keyword arguments passed to `__init__` method.

        :keyword int limit: limit for memory usage by cached items in megabytes.
                            if not provided, it will be get from
                            `caching` config store. if you want to
                            remove size limit, you could pass
                            `caching.globals.NO_LIMIT` as input. note
                            that this will remove limit only by client,
                            so memcached server default limit will be intact.

        :keyword str hostname: the host name of memcached server. it could be
                               ip, url or `localhost`.

        :keyword int port: port number of memcached server.

        :keyword str unixsocket: path to unix socket file.

        :note hostname, port and unixsocket: both unixsocket and hostname, port
                                             could not be provided at the same time.

        :keyword int connect_timeout: milliseconds to wait for a connection to
                                      the memcached server. defaults to `forever`
                                      (uses the underlying default socket timeout,
                                      which can be very long).

        :keyword int timeout: milliseconds to wait for send or recv calls on
                              the socket connected to memcached. defaults to
                              `forever` (uses the underlying default socket
                              timeout, which can be very long).

        :keyword bool no_delay: set the TCP_NODELAY flag, which may help with
                                performance in some cases. defaults to False.

        :keyword bool ignore_exc: True to cause the `get`, `gets`, `get_many` and
                                  `gets_many` calls to treat any errors as cache
                                  misses. defaults to False.

        :keyword bool default_noreply: the default value for `noreply` as passed to
                                       store commands (except from cas, incr, and decr,
                                       which default to False).

        :keyword bool allow_unicode_keys: support unicode (utf8) keys.
        :keyword str encoding: controls data encoding. defaults to `ascii`.

        :raises InvalidCacheLimitError: invalid cache limit error.

        :raises BothHostnameAndUnixSocketProvidedError: both hostname and unix
                                                        socket provided error.

        :raises HostnameOrUnixSocketRequiredError: hostname or unix socket
                                                   required error.

        :rtype: pymemcache.Client
        """

        if kwargs is None:
            kwargs = {}

        limit = kwargs.get('limit') or configs.pop('limit', None)
        if limit is None or (limit != NO_LIMIT and limit <= 0):
            raise InvalidCacheLimitError('Cache limit for cache [{name}] '
                                         'must be a positive integer.'
                                         .format(name=self.get_name()))
        self._limit = limit

        # we should remove 'limit' from configs to be able to pass it to Client initializer.
        configs.pop('limit', None)

        unixsocket = configs.pop('unixsocket', None)
        hostname = configs.pop('hostname', None)
        port = configs.pop('port', None)

        if unixsocket is not None and (hostname is not None or port is not None):
            raise BothHostnameAndUnixSocketProvidedError('Both hostname and unix socket could '
                                                         'not be provided at the same time.')

        if unixsocket is None and (hostname is None or port is None):
            raise HostnameOrUnixSocketRequiredError('Hostname or unix socket must be provided '
                                                    'to connect to memcached server.')

        server = None
        if unixsocket is not None:
            server = unixsocket
        else:
            server = hostname, port

        timeout = configs.get('timeout')
        if timeout is not None and timeout > 0:
            timeout = timeout / 1000
            configs.update(timeout=timeout)

        connect_timeout = configs.get('connect_timeout')
        if connect_timeout is not None and connect_timeout > 0:
            connect_timeout = connect_timeout / 1000
            configs.update(connect_timeout=connect_timeout)

        configs.update(serde=self)
        client = Client(server, **configs)
        if self.limit != NO_LIMIT:
            client.cache_memlimit(self.limit)

        return client

    def _increase_hit(self):
        """
        increases hit count of this cache.
        """

        self.client.incr(self.HIT_KEY, 1, noreply=True)

    def _increase_miss(self):
        """
        increases miss count of this cache.
        """

        self.client.incr(self.MISS_KEY, 1, noreply=True)

    def _clear(self):
        """
        clears all items from cache.
        """

        self.client.flush_all(noreply=True)

    def _set(self, key, value, *args, **options):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :keyword int expire: expire time for this item in seconds.
                             if not provided, it will be get from `expire` attribute.

        :keyword bool noreply: True to not wait for the reply. if not
                               provided, defaults to `caching` config store.

        :keyword int flags: arbitrary bit field used for memcached server-specific flags.
        """

        expire = options.get('expire')
        if expire is None:
            expire = self.expire

        self.client.set(key, value, expire=expire,
                        noreply=options.get('noreply'),
                        flags=options.get('flags'))

    def _get(self, key, default=None, **options):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        result = self.client.get(key, default=default)
        if result in (None, default):
            self._increase_miss()
            return result

        self._increase_hit()
        return result

    def _prepare_key(self, key):
        """
        prepares the key to be used in cache.

        :param object key: key to be cached.

        :returns: prepared key.
        :rtype: str
        """

        return str(key)

    def serialize(self, key, value):
        """
        serializes the given value.

        :param object key: key of value to be serialized.
        :param object value: value to be serialized.

        :returns: serialized value
        :rtype: tuple[bytes | int, int]
        """

        if isinstance(value, int):
            return value, 1

        return pickle.dumps(value), 2

    def deserialize(self, key, value, flags=None):
        """
        deserializes the given value.

        :param object key: key of value to be deserialized.
        :param bytes value: value to be deserialized.

        :param int flags: bitwise flags. if you want to deserialize
                          an integer value, you could set it to 1.

        :returns: deserialized value
        """

        if flags == 1:
            return int(value)

        return pickle.loads(value)

    def increment(self, key, value, noreply=False):
        """
        increments the value of given key by provided value.

        the cached and provided values must be integers.
        it returns the new value if key is existed, otherwise returns None.
        note that if noreply=True, it always returns None.

        :param object key: hashable key to increase its value in the cache.
        :param int value: the amount by which to increment the cached value.
        :param bool noreply: True to not wait for the reply. defaults to False.

        :rtype: int
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        return self.client.incr(hashed_key, value, noreply)

    def decrement(self, key, value, noreply=False):
        """
        decrements the value of given key by provided value.

        the cached and provided values must be integers.
        it returns the new value if key is existed, otherwise returns None.
        note that if noreply=True, it always returns None.

        :param object key: hashable key to decrease its value in the cache.
        :param int value: the amount by which to decrement the cached value.
        :param bool noreply: True to not wait for the reply. defaults to False.

        :rtype: int
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        return self.client.decr(hashed_key, value, noreply)

    def contains(self, key):
        """
        gets a value indicating that given key is existed in the cached items.

        :param object key: key to be checked for existence.

        :rtype: bool
        """

        result = self.get(key, default=None)
        return result is not None

    def pop(self, key, default=None):
        """
        pops the given key from cache and returns its value.

        if key does not exist, it returns None or the specified default value.

        :param object key: key to get its value.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        result = self.get(key, default=default)
        self.remove(key, noreply=True)
        return result

    def remove(self, key, **options):
        """
        removes the given key from cache.

        it does nothing if the key is not present in the cache.

        :keyword bool noreply: do not wait for reply from server.
                               will be get from `caching` config store
                               if not provided.

        :param object key: key to be removed.
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        self.client.delete(hashed_key, noreply=options.get('noreply'))

    def touch(self, key, expire=0, noreply=None):
        """
        extends the expire time of given key in the cache.


        :param object key: hashable key to extend its expire time.

        :param int expire: expire time for this item in seconds.
                           if not provided, it will be get from `expire` attribute.

        :param bool noreply: True to not wait for the reply. if not
                             provided, defaults to `caching` config store.

        :returns: True if expire time updated, False if the key is not existed.
        :rtype: bool
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)

        if expire is None:
            expire = self.expire

        return self.client.touch(hashed_key, expire=expire, noreply=noreply)

    def add(self, key, value, expire=0, noreply=None, flags=None):
        """
        adds the given key into the cache if it isn't already, otherwise does nothing.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :param int expire: expire time for this item in seconds.
                           if not provided, it will be get from `expire` attribute.

        :param bool noreply: True to not wait for the reply. if not
                             provided, defaults to `caching` config store.

        :param int flags: arbitrary bit field used for memcached server-specific flags.

        :returns: True if value is set, False if the key is already existed.
        :rtype: bool
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)

        if expire is None:
            expire = self.expire

        return self.client.add(hashed_key, value, expire=expire, noreply=noreply, flags=flags)

    @property
    def stats(self):
        """
        get the statistic info about cached items.

        :returns: dict(datetime last_cleared_time: last cleared time,
                       bool persistent: persistent cache,
                       bool consider_user: consider user,
                       int expire: cached items expire time,
                       int hit: hit count,
                       int miss: miss count,
                       str hit_ratio: hit ratio,
                       int limit: memory size limit)
        :rtype: dict
        """

        base_stats = super().stats
        stats = dict(limit=self.limit)
        base_stats.update(stats)
        return base_stats

    @property
    def hit_count(self):
        """
        gets the hit count for this cache.

        :rtype: int
        """

        return self.client.get(self.HIT_KEY, default=0)

    @property
    def miss_count(self):
        """
        gets the miss count for this cache.

        :rtype: int
        """

        return self.client.get(self.MISS_KEY, default=0)

    @property
    def limit(self):
        """
        gets the memory size limit of this cache in megabytes.

        :rtype: int
        """

        return self._limit



# -*- coding: utf-8 -*-
"""
validator decorators module.
"""

import pyrin.validator.services as validator_services


def validator(*args, **kwargs):
    """
    decorator to register a validator.

    :param object args: validator class constructor arguments.
    :param object kwargs: validator class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           validator with the same name and domain, replace
                           it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidValidatorTypeError: invalid validator type error.
    :raises DuplicatedValidatorError: duplicated validator error.

    :returns: validator class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available validator.

        :param type cls: validator class.

        :returns: validator class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        validator_services.register_validator(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
packaging hooks module.
"""

from pyrin.core.structs import Hook


class PackagingHookBase(Hook):
    """
    packaging hook base class.

    all packages that need to be hooked into packaging business must
    implement this class and register it in packaging hooks.
    """

    def after_packages_loaded(self):
        """
        this method will be called after all application packages have been loaded.
        """
        pass

    def package_loaded(self, package_name, **options):
        """
        this method will be called after each application package has been loaded.

        :param str package_name: name of the loaded package.
        """
        pass



# -*- coding: utf-8 -*-
"""
accounting exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class AccountingException(CoreException):
    """
    accounting exception.
    """
    pass


class AccountingBusinessException(CoreBusinessException, AccountingException):
    """
    accounting business exception.
    """
    pass



# -*- coding: utf-8 -*-
"""
core decorators module.
"""


class class_property(object):
    """
    a decorator to define a class property.

    it is similar to a property but works at class level.
    it only supports get.

    usage example:

    @class_property
    def name(cls):
        return cls.VALUE
    """

    def __init__(self, method=None):
        """
        initializes an instance of class_property.

        :param function method: decorated method.
        """

        self.fget = method

    def __get__(self, instance, cls=None):
        """
        gets the result of decorated method.

        :param instance: instance of parent class.
        :param type cls: class type.

        :returns: decorated method result.
        """

        if cls is None:
            cls = type(instance)

        return self.fget(cls)

    def getter(self, method):
        """
        returns current instance with its getter set to given method.

        :param function method: decorated method.

        :rtype: class_property
        """

        self.fget = method
        return self



# -*- coding: utf-8 -*-
"""
permission models module.
"""

from sqlalchemy import Unicode, SmallInteger

from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.sql.schema.base import CoreColumn


class PermissionBaseEntity(CoreEntity):
    """
    permission base entity class.
    """

    _table = 'permission'

    id = CoreColumn(name='id', type_=SmallInteger, primary_key=True, nullable=False, index=True)


class PermissionEntity(PermissionBaseEntity):
    """
    permission entity class.
    """

    _extend_existing = True

    description = CoreColumn(name='description', type_=Unicode(100), nullable=False)



# -*- coding: utf-8 -*-
"""
users services module.
"""

from pyrin.application.services import get_component

from demo.security.users import UsersPackage


def create(username, password, first_name, last_name, **options):
    """
    creates a new user based on given inputs.

    :param str username: username.
    :param str password: password.
    :param str first_name: first name.
    :param str last_name: last name.
    """

    get_component(UsersPackage.COMPONENT_NAME).create(username, password, first_name,
                                                      last_name, **options)


def get(**options):
    """
    gets the current user info.

    :raises UserNotFoundError: user not found error.

    :rtype: UserEntity
    """

    return get_component(UsersPackage.COMPONENT_NAME).get(**options)


def get_all(**options):
    """
    gets a list of all users.

    :returns: list[UserEntity]
    :rtype: list
    """

    return get_component(UsersPackage.COMPONENT_NAME).get_all(**options)



# -*- coding: utf-8 -*-
"""
validator handlers string module.
"""

import re

from pyrin.core.globals import _
from pyrin.validator.handlers.base import ValidatorBase
from pyrin.validator.handlers.exceptions import LongStringLengthError, ShortStringLengthError, \
    ValueCouldNotBeBlankError, ValueCouldNotBeWhitespaceError, ValueDoesNotMatchPatternError, \
    InvalidRegularExpressionError, RegularExpressionMustBeProvidedError, ValueIsNotStringError, \
    MinimumLengthHigherThanMaximumLengthError, InvalidEmailError, InvalidIPv4Error, \
    InvalidURLError, InvalidHTTPURLError, InvalidHTTPSURLError


class StringValidator(ValidatorBase):
    """
    string validator class.
    """

    invalid_type_error = ValueIsNotStringError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be a string.')
    long_length_error = LongStringLengthError
    long_length_message = _('The provided value for [{param_name}] '
                            'should have at most [{count}] characters.')
    short_length_error = ShortStringLengthError
    short_length_message = _('The provided value for [{param_name}] '
                             'should have at least [{count}] characters.')
    blank_value_error = ValueCouldNotBeBlankError
    blank_value_message = _('The provided value for [{param_name}] '
                            'could not be blank.')
    whitespace_value_error = ValueCouldNotBeWhitespaceError
    whitespace_value_message = _('The provided value for [{param_name}] '
                                 'could not be whitespace.')

    default_allow_blank = None
    default_allow_whitespace = None
    default_minimum_length = None
    default_maximum_length = None

    def __init__(self, domain, name, **options):
        """
        initializes an instance of StringValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_blank: specifies that empty strings should be accepted
                                   as valid. defaults to True if not provided.

        :keyword bool allow_whitespace: specifies that whitespace strings should be accepted
                                        as valid. defaults to False if not provided.

        :keyword int minimum_length: specifies the minimum valid length for string value.
                                     no min length checking will be done if not provided.

        :keyword int maximum_length: specifies the maximum valid length for string value.
                                     no max length checking will be done if not provided.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        :raises LongStringLengthError: long string length error.
        :raises ShortStringLengthError: short string length error.
        :raises ValueCouldNotBeBlankError: value could not be blank error.
        :raises ValueCouldNotBeWhitespaceError: value could not be whitespace error.
        :raises MinimumLengthHigherThanMaximumLengthError: minimum length higher than
                                                           maximum length error.
        """

        options.update(accepted_type=str)
        super().__init__(domain, name, **options)

        allow_blank = options.get('allow_blank')
        if allow_blank is None:
            if self.default_allow_blank is not None:
                allow_blank = self.default_allow_blank
            else:
                allow_blank = True

        allow_whitespace = options.get('allow_whitespace')
        if allow_whitespace is None:
            if self.default_allow_whitespace is not None:
                allow_whitespace = self.default_allow_whitespace
            else:
                allow_whitespace = False

        self._minimum_length = options.get('minimum_length')
        if self._minimum_length is None:
            self._minimum_length = self.default_minimum_length

        self._maximum_length = options.get('maximum_length')
        if self._maximum_length is None:
            self._maximum_length = self.default_maximum_length

        if self._minimum_length is not None and self._maximum_length is not None \
                and self._minimum_length > self._maximum_length:
            raise MinimumLengthHigherThanMaximumLengthError('Minimum length of string '
                                                            'could not be higher than '
                                                            'maximum length.')

        self._allow_blank = allow_blank
        self._allow_whitespace = allow_whitespace

        self._validate_exception_type(self.long_length_error)
        self._validate_exception_type(self.short_length_error)
        self._validate_exception_type(self.blank_value_error)
        self._validate_exception_type(self.whitespace_value_error)

    def _validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate()`
        preferably at the beginning.

        :param str value: value to be validated.

        :keyword bool allow_blank: determines that empty strings should be
                                   considered valid. this value has precedence
                                   over `allow_blank` instance attribute
                                   if provided.

        :keyword bool allow_whitespace: determines that whitespace strings should be
                                        considered valid. this value has precedence
                                        over `allow_whitespace` instance attribute
                                        if provided.

        :raises LongStringLengthError: long string length error.
        :raises ShortStringLengthError: short string length error.
        :raises ValueCouldNotBeBlankError: value could not be blank error.
        :raises ValueCouldNotBeWhitespaceError: value could not be whitespace error.
        """

        super()._validate(value, **options)

        allow_blank = options.get('allow_blank')
        if allow_blank is None:
            allow_blank = self.allow_blank

        allow_whitespace = options.get('allow_whitespace')
        if allow_whitespace is None:
            allow_whitespace = self.allow_whitespace

        length = len(value)
        if self.maximum_length is not None and length > self.maximum_length:
            raise self.long_length_error(self.long_length_message.format(
                param_name=self.localized_name, count=self.maximum_length))

        if self.minimum_length is not None and length < self.minimum_length:
            raise self.short_length_error(self.short_length_message.format(
                param_name=self.localized_name, count=self.minimum_length))

        if allow_blank is not True and length == 0:
            raise self.blank_value_error(self.blank_value_message.format(
                param_name=self.localized_name))

        if allow_whitespace is not True and value.isspace():
            raise self.whitespace_value_error(self.whitespace_value_message.format(
                param_name=self.localized_name))

        if length > 0 and not value.isspace():
            self._validate_extra(value, **options)

    def _validate_extra(self, value, **options):
        """
        validates the given value.

        this method is intended to be overridden by subclasses.
        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate_extra()`
        preferably at the beginning.

        :param str value: value to be validated.

        :raises ValidationError: validation error.
        """
        pass

    @property
    def minimum_length(self):
        """
        gets the minimum accepted length of this validator.

        returns None if no minimum has been set.

        :rtype: int
        """

        return self._minimum_length

    @property
    def maximum_length(self):
        """
        gets the maximum accepted length of this validator.

        returns None if no maximum has been set.

        :rtype: int
        """

        return self._maximum_length

    @property
    def allow_blank(self):
        """
        gets a value indicating that empty strings are considered valid.

        :rtype: bool
        """

        return self._allow_blank

    @property
    def allow_whitespace(self):
        """
        gets a value indicating that whitespace strings are considered valid.

        :rtype: bool
        """

        return self._allow_whitespace


class RegexValidator(StringValidator):
    """
    regex validator class.
    """

    # the regular expression to use for validation.
    # it could be set with a string containing a regex
    # or with a `Pattern` object from re.compile() method.
    regex = None
    pattern_not_match_error = ValueDoesNotMatchPatternError
    pattern_not_match_message = _('The provided value for [{param_name}] '
                                  'does not match the required pattern.')

    default_flags = re.IGNORECASE | re.DOTALL

    def __init__(self, domain, name, **options):
        """
        initializes an instance of RegexValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_blank: specifies that empty strings should be accepted
                                   as valid. defaults to True if not provided.

        :keyword bool allow_whitespace: specifies that whitespace strings should be accepted
                                        as valid. defaults to False if not provided.

        :keyword int minimum_length: specifies the minimum valid length for string value.
                                     no min length checking will be done if not provided.

        :keyword int maximum_length: specifies the maximum valid length for string value.
                                     no max length checking will be done if not provided.

        :keyword int flags: flags to be used for regular expression
                            compilation using `re.compile` method.
                            this will only be used if a string regex is provided.
                            if no flags are provided, `default_flags` will be applied.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        :raises LongStringLengthError: long string length error.
        :raises ShortStringLengthError: short string length error.
        :raises ValueCouldNotBeBlankError: value could not be blank error.
        :raises ValueCouldNotBeWhitespaceError: value could not be whitespace error.
        :raises InvalidRegularExpressionError: invalid regular expression error.
        :raises RegularExpressionMustBeProvidedError: regular expression must be provided error.
        """

        super().__init__(domain, name, **options)

        is_string = isinstance(self.regex, str)
        if not isinstance(self.regex, re.Pattern) and not is_string:
            raise InvalidRegularExpressionError('The provided regular expression is invalid. '
                                                'it must be a string containing a regular '
                                                'expression or a Pattern object returned from '
                                                're.compile() method.')

        if is_string and (len(self.regex) <= 0 or self.regex.isspace()):
            raise RegularExpressionMustBeProvidedError('The provided string for regular '
                                                       'expression could not be blank.')

        if is_string:
            flags = options.get('flags')
            if flags is None:
                flags = self.default_flags
            self._pattern = re.compile(self.regex, flags=flags)
        else:
            self._pattern = self.regex

        self._validate_exception_type(self.pattern_not_match_error)

    def _validate_extra(self, value, **options):
        """
        validates the given value.

        this method is intended to be overridden by subclasses.
        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate_extra()`
        preferably at the beginning.

        :param str value: value to be validated.

        :raises ValueDoesNotMatchPatternError: value does not match pattern error.
        """

        super()._validate_extra(value, **options)

        if not self.pattern.match(value):
            raise self.pattern_not_match_error(
                self.pattern_not_match_message.format(param_name=self.localized_name))

    @property
    def pattern(self):
        """
        gets the pattern of this validator.

        :rtype: re.Pattern
        """

        return self._pattern


class EmailValidator(RegexValidator):
    """
    email validator class.
    """

    regex = r'^[a-z0-9]+([a-z0-9\.]*[a-z0-9]+)*[@]\w+[\.]\w{2,3}([\.]\w{2,3})?$'
    pattern_not_match_error = InvalidEmailError
    pattern_not_match_message = _('The provided value for [{param_name}] '
                                  'is not a valid email address.')

    default_minimum_length = 6

    def _validate_extra(self, value, **options):
        """
        validates the given value.

        this method is intended to be overridden by subclasses.
        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate_extra()`
        preferably at the beginning.

        :param str value: value to be validated.

        :raises InvalidEmailError: invalid email error.
        """

        super()._validate_extra(value, **options)

        if '..' in value:
            raise self.pattern_not_match_error(
                self.pattern_not_match_message.format(param_name=self.localized_name))


class IPv4Validator(RegexValidator):
    """
    ipv4 validator class.
    """

    regex = r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$'

    pattern_not_match_error = InvalidIPv4Error
    pattern_not_match_message = _('The provided value for [{param_name}] '
                                  'is not a valid IPv4 address.')

    default_maximum_length = 15
    default_minimum_length = 7

    def _validate_extra(self, value, **options):
        """
        validates the given value.

        this method is intended to be overridden by subclasses.
        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate_extra()`
        preferably at the beginning.

        :param str value: value to be validated.

        :raises InvalidIPv4Error: invalid ipv4 error.
        """

        super()._validate_extra(value, **options)

        parts = value.split('.')
        for item in parts:
            converted_item = int(item)
            if converted_item < 0 or converted_item > 255:
                raise self.pattern_not_match_error(
                    self.pattern_not_match_message.format(param_name=self.localized_name))


class URLValidator(RegexValidator):
    """
    url validator class.

    this matches urls starting with `www`.
    """

    regex = r'^www\..+\.\w+$'

    pattern_not_match_error = InvalidURLError
    pattern_not_match_message = _('The provided value for [{param_name}] '
                                  'is not a valid url.')

    default_minimum_length = 7

    def _validate_extra(self, value, **options):
        """
        validates the given value.

        this method is intended to be overridden by subclasses.
        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        each overridden method must call `super()._validate_extra()`
        preferably at the beginning.

        :param str value: value to be validated.

        :raises InvalidURLError: invalid url error.
        """

        super()._validate_extra(value, **options)

        if ' ' in value or '..' in value:
            raise self.pattern_not_match_error(
                self.pattern_not_match_message.format(param_name=self.localized_name))


class HTTPValidator(URLValidator):
    """
    http validator class.

    this matches urls starting with `http://www`.
    """

    regex = r'^http://www\..+\.\w+$'

    pattern_not_match_error = InvalidHTTPURLError
    pattern_not_match_message = _('The provided value for [{param_name}] '
                                  'is not a valid http url.')

    default_minimum_length = 14


class HTTPSValidator(URLValidator):
    """
    https validator class.

    this matches urls starting with `https://www`.
    """

    regex = r'^https://www\..+\.\w+$'

    pattern_not_match_error = InvalidHTTPSURLError
    pattern_not_match_message = _('The provided value for [{param_name}] '
                                  'is not a valid https url.')

    default_minimum_length = 15



# -*- coding: utf-8 -*-
"""
database audit manager module.
"""

import traceback

import pyrin.database.services as database_services

from pyrin.audit.enumerations import InspectionStatusEnum
from pyrin.database.audit import DatabaseAuditPackage
from pyrin.core.structs import Manager
from pyrin.database.services import get_current_store


class DatabaseAuditManager(Manager):
    """
    database audit manager class.
    """

    package_class = DatabaseAuditPackage

    def inspect(self, **options):
        """
        inspects the status of available databases.

        it returns a tuple of two values. first value is a dict containing the inspection
        data. and the second value is a bool value indicating that inspection has been
        succeeded or failed.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :rtype: tuple[dict, bool]
        """

        all_databases = database_services.get_all_database_names()
        default_database = database_services.get_default_database_name()
        all_succeeded = True
        data = {}

        for item in all_databases:
            name = item
            if item == default_database:
                name = None
            inspection_data, succeeded = self._test_database(name, **options)
            if succeeded is False:
                all_succeeded = False
            data[item] = inspection_data

        return data, all_succeeded

    def _test_database(self, bind_name=None, **options):
        """
        tests that given bind name database is ready.

        it returns a tuple of two values. first value is a dict containing the inspection
        data. and the second value is a bool value indicating that inspection has been
        succeeded or failed.

        :param str bind_name: database bind name from `database.binds` config store.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :rtype: tuple[dict, bool]
        """

        raise_error = options.get('raise_error', False)
        include_traceback = options.get('traceback', True)
        store = get_current_store()
        data = {}
        succeeded = True
        try:
            store.execute('select 1', bind_name=bind_name)
            data.update(status=InspectionStatusEnum.OK)
        except Exception as error:
            if raise_error is True:
                raise

            succeeded = False
            data.update(status=InspectionStatusEnum.FAILED,
                        error=str(error))
            if include_traceback is not False:
                data.update(traceback=traceback.format_exc())

        return data, succeeded



# -*- coding: utf-8 -*-
"""
orm sql schema module.
"""

from sqlalchemy import Column

from pyrin.caching.decorators import cached_property
from pyrin.database.orm.sql.operators.base import CoreColumnOperators


class CoreColumn(Column, CoreColumnOperators):
    """
    core column class.

    all application models columns must be an instance of this class.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :param bool autoincrement: set up `auto increment` semantics for an
                                   integer primary key column.

        :param callable | object default: a scalar, python callable or `ColumnElement`
                                          expression representing the default value
                                          for this column, which will be invoked upon
                                          insert if this column is otherwise not
                                          specified in the values clause of the insert.

        :param str doc: optional string that can be used by the ORM or similar
                        to document attributes on the python side.

        :param str key: an optional string identifier which will identify this
                        `Column` object on the `Table`.

        :param bool index: When `True`, indicates that the column is indexed.

        :param dict info: optional data dictionary which will be populated into the
                          `SchemaItem.info` attribute of this object.

        :param bool nullable: when set to `False`, will cause the `Not NULL`
                              phrase to be added when generating ddl for the column.

        :param callable | object onupdate: a scalar, python callable, or
                                           `ClauseElement` representing a default
                                           value to be applied to the column within update
                                           statements, which will be invoked upon update
                                           if this column is not present in the set
                                           clause of the update.

        :param bool primary_key: if `True`, marks this column as a primary key
                                 column. multiple columns can have this flag set to
                                 specify composite primary keys.

        :param object server_default: a `FetchedValue` instance, str, unicode
                                      or `text` construct representing the ddl
                                      default value for the column.

        :param FetchedValue server_onupdate: a `FetchedValue` instance representing a
                                             database-side default generation function,
                                             such as a trigger. this indicates to sqlalchemy
                                             that a newly generated value will be available
                                             after updates. this construct does not actually
                                             implement any kind of generation function within
                                             the database, which instead must be specified
                                             separately.

        :param bool quote: force quoting of this column's name on or off,
                           corresponding to `True` or `False`. when left at its default
                           of `None`, the column identifier will be quoted according to
                           whether the name is case sensitive (identifiers with at least one
                           upper case character are treated as case sensitive), or if it's a
                           reserved word.

        :param bool unique: when `True`, indicates that this column contains a
                            unique constraint, or if `index` is `True` as well, indicates
                            that the `index` should be created with the unique flag.

        :param bool system: when `True`, indicates this is a system column,
                            that is a column which is automatically made available by the
                            database, and should not be included in the columns list for a
                            `create table` statement.

        :param str comment: optional string that will render an sql comment
                            on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to True if not provided.
        """

        self.allow_read = kwargs.pop('allow_read', True)
        self.allow_write = kwargs.pop('allow_write', True)

        super().__init__(*args, **kwargs)

    @cached_property
    def fullname(self):
        """
        gets the column's fullname.

        fullname is made up of `table_fullname.column_name`.
        if the column has no table, it only returns the `column_name`.

        :rtype: str
        """

        if self.table is None:
            return self._real_name()

        for single_column in self.table.columns:
            if single_column.name == self.name:
                for base_column in single_column.base_columns:
                    return '{table}.{column}'.format(table=base_column.table.fullname,
                                                     column=self._real_name())

    @property
    def is_foreign_key(self):
        """
        gets a value indicating that this column is a foreign key.

        :rtype: bool
        """

        return len(self.foreign_keys) > 0

    def _real_name(self):
        """
        gets the column's real name.

        :rtype: str
        """

        for column in self.base_columns:
            return column.name



# -*- coding: utf-8 -*-
"""
string normalizer interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class StringNormalizerSingletonMeta(MultiSingletonMeta):
    """
    string normalizer singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractStringNormalizerBase(CoreObject, metaclass=StringNormalizerSingletonMeta):
    """
    abstract string normalizer base class.
    """

    @abstractmethod
    def normalize(self, value, *args, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :keyword bool strip: strip spaces from both ends of value.
                             defaults to True if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: normalized value.
        :rtype: str
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def priority(self):
        """
        gets the priority of this normalizer.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: int
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
cli core template handlers module.
"""

import os

import pyrin.utils.path as path_utils

from pyrin.core.structs import DTO
from pyrin.template.handlers.base import TemplateHandlerWithInterfaceInputBase
from pyrin.cli.core.template.exceptions import InvalidProjectRootPathError, \
    InvalidApplicationPackageNameError, InvalidApplicationClassNameError, \
    InvalidProjectStructureTemplateHandlerType


class ProjectStructureTemplateHandlerBase(TemplateHandlerWithInterfaceInputBase):
    """
    project structure template handler base class.
    """

    def __init__(self, name, source):
        """
        initializes an instance of ProjectStructureTemplateHandlerBase.

        :param str name: name of the handler.
                         each handler must have a unique name.

        :param str source: source directory of template files.
        """

        self._next_handler = None
        self._app_package_name = None
        self._app_class_name = None
        self._project_root = None

        super().__init__(name, source)

    def set_next(self, instance):
        """
        sets the next template handler.

        :param ProjectStructureTemplateHandlerBase instance: template handler instance.

        :raises InvalidProjectStructureTemplateHandlerType: invalid project structure
                                                            template handler type.

        :rtype: ProjectStructureTemplateHandlerBase
        """

        if not isinstance(instance, ProjectStructureTemplateHandlerBase):
            raise InvalidProjectStructureTemplateHandlerType('Input parameter [{instance}] '
                                                             'is not an instance of [{base}]'
                                                             .format(instance=instance,
                                                                     base=self))

        self._next_handler = instance
        return instance

    def _show_interface(self, project_root=None,
                        app_package_name=None, app_class_name=None):
        """
        shows cli prompt to get inputs from user.

        :param str project_root: project root directory name or full path.
                                 if only a directory name is given, it will
                                 be created in current path. defaults to None
                                 if not provided and will be get from command prompt.

        :param str app_package_name: application package name.
                                     defaults to None if not provided and
                                     will be get it from command prompt.

        :param str app_class_name: application class name.
                                   defaults to None if not provided and
                                   will be get it from command prompt.

        :raises InvalidProjectRootPathError: invalid project root path error.
        :raises InvalidApplicationPackageNameError: invalid application package name error.
        :raises InvalidApplicationClassNameError: invalid application class name error.
        """

        if project_root is None:
            project_root = input('Please input the project path or directory name: ')

        if app_package_name is None:
            app_package_name = input('Please input the package name of your application: ')

        if app_class_name is None:
            app_class_name = input('Please input your application class name: ')

        self._set_attributes(project_root, app_package_name, app_class_name)

    def _validate_inputs(self, project_root, app_package_name, app_class_name):
        """
        validates the inputs to be used by this handler.

        :param str project_root: project root directory name or full path.
                                 if only a directory name is given, it will
                                 be created in current path.

        :param str app_package_name: application package name.
        :param str app_class_name: application class name.

        :raises InvalidProjectRootPathError: invalid project root path error.
        :raises InvalidApplicationPackageNameError: invalid application package name error.
        :raises InvalidApplicationClassNameError: invalid application class name error.
        """

        if project_root in (None, '') or project_root.isspace() or \
                not os.path.isabs(project_root):
            raise InvalidProjectRootPathError('Project root path is invalid.')

        if app_package_name in (None, '') or app_package_name.isspace():
            raise InvalidApplicationPackageNameError('Application package name is invalid.')

        if app_class_name in (None, '') or app_class_name.isspace():
            raise InvalidApplicationClassNameError('Application class name is invalid.')

    def _set_attributes(self, project_root, app_package_name, app_class_name):
        """
        sets the required attributes based on given inputs.

        :param str project_root: project root directory name or full path.
                                 if only a directory name is given, it will
                                 be created in current path.

        :param str app_package_name: application package name.
        :param str app_class_name: application class name.

        :raises InvalidProjectRootPathError: invalid project root path error.
        :raises InvalidApplicationPackageNameError: invalid application package name error.
        :raises InvalidApplicationClassNameError: invalid application class name error.
        """

        if project_root not in (None, '') and \
                not project_root.isspace() and not os.path.isabs(project_root):
            project_root = os.path.abspath(os.path.join('.', project_root))

        self._validate_inputs(project_root, app_package_name, app_class_name)

        self._app_class_name = app_class_name.replace(' ', '')
        self._app_package_name = app_package_name.lower().replace(' ', '')
        self._project_root = project_root.rstrip(os.path.sep).rstrip(os.path.altsep)
        self._target = self._process_target(self._project_root, self._app_package_name)

    def _process_target(self, project_root, app_package_name):
        """
        processes and gets the target value to be set for this handler's target attribute.

        it gets the project root value as target, but it could be overridden
        in subclasses to customize the target value based on given inputs.

        :param str project_root: project root absolute path.
        :param str app_package_name: application package name.
        """

        return project_root

    def _get_file_patterns(self):
        """
        gets the file patterns that should be included in replacement operation.

        :rtype: list[str]
        """

        return ['.py']

    def _get_data(self):
        """
        gets the data required in template generation to replace in files.

        :rtype: dict
        """

        return DTO(APPLICATION_PACKAGE=self._app_package_name,
                   APPLICATION_CLASS=self._app_class_name)

    def _finalize(self):
        """
        finalizes the template creation.

        it actually calls the next handler if available, providing
        it the inputs to prevent getting them from the user again.
        """

        if self._next_handler is not None:
            return self._next_handler.create(self._project_root,
                                             self._app_package_name,
                                             self._app_class_name)


class ScriptsTemplateHandler(ProjectStructureTemplateHandlerBase):
    """
    scripts template handler class.
    """

    def __init__(self):
        """
        initializes an instance of ScriptsTemplateHandler.
        """

        pyrin_path = path_utils.get_main_package_path(self.__module__)
        source = os.path.abspath(os.path.join(pyrin_path, 'cli', 'core',
                                              'template', 'files', 'scripts'))

        super().__init__('scripts', source)


class ApplicationTemplateHandler(ProjectStructureTemplateHandlerBase):
    """
    application template handler class.
    """

    def __init__(self):
        """
        initializes an instance of ApplicationTemplateHandler.
        """

        pyrin_path = path_utils.get_main_package_path(self.__module__)
        source = os.path.abspath(os.path.join(pyrin_path, 'cli', 'core',
                                              'template', 'files', 'application'))

        super().__init__('application', source)

    def _process_target(self, project_root, app_package_name):
        """
        processes and gets the target value to be set for this handler's target attribute.

        it gets the application's main package path as target.

        :param str project_root: project root absolute path.
        :param str app_package_name: application package name.
        """

        return os.path.abspath(os.path.join(project_root, app_package_name))



# -*- coding: utf-8 -*-
"""
cors hooks module.
"""

import pyrin.processor.cors.services as cors_services

from pyrin.application.decorators import application_hook
from pyrin.application.hooks import ApplicationHookBase


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def provide_response_headers(self, headers, endpoint,
                                 status_code, method, **options):
        """
        this method will be called whenever a response is going to be returned from server.

        :param dict | Headers headers: current response headers.

        :param str endpoint: the endpoint of the route that
                             handled the current request.
                             by default, it is the fully qualified
                             name of the view function.

        :param int status_code: response status code.
                                it could be None if not provided.

        :param str method: the http method of current request.

        :keyword str url: the url of the route that handled this request.

        :keyword user: the user of current request.
                       it could be None.
        """

        cors_headers = cors_services.get_current_cors_headers()
        if cors_headers is not None:
            headers.extend(cors_headers)



# -*- coding: utf-8 -*-
"""
application exceptions module.
"""

from pyrin.core.exceptions import ContextAttributeError, CoreException


class ApplicationException(CoreException):
    """
    application exception.
    """
    pass


class ComponentAttributeError(ApplicationException, ContextAttributeError):
    """
    component attribute error.
    """
    pass


class InvalidComponentTypeError(ApplicationException):
    """
    invalid component type error.
    """
    pass


class InvalidComponentIDError(ApplicationException):
    """
    invalid component id error.
    """
    pass


class InvalidComponentNameError(ApplicationException):
    """
    invalid component name error.
    """
    pass


class DuplicateComponentIDError(ApplicationException):
    """
    duplicate component id error.
    """
    pass


class DuplicateContextKeyError(ApplicationException):
    """
    duplicate context key error.
    """
    pass


class DuplicateRouteURLError(ApplicationException):
    """
    duplicate route url error.
    """
    pass


class InvalidRouteFactoryTypeError(ApplicationException):
    """
    invalid route factory type error.
    """
    pass


class InvalidApplicationStatusError(ApplicationException):
    """
    invalid application status error.
    """
    pass


class ApplicationInScriptingModeError(ApplicationException):
    """
    application in scripting mode error.
    """
    pass


class ApplicationInstanceAlreadySetError(ApplicationException):
    """
    application instance already set error.
    """
    pass


class ApplicationIsNotSubclassedError(ApplicationException):
    """
    application is not subclassed error.
    """
    pass


class InvalidApplicationHookTypeError(ApplicationException):
    """
    invalid application hook type error.
    """
    pass


class OverwritingEndpointIsNotAllowedError(ApplicationException):
    """
    overwriting endpoint is not allowed error.
    """
    pass


class InvalidStaticHostAndHostMatchingError(ApplicationException):
    """
    invalid static host and host matching error.
    """
    pass



# -*- coding: utf-8 -*-
"""
sentry decorators module.
"""

import pyrin.logging.sentry.services as sentry_services


def sentry_integration(*args, **kwargs):
    """
    decorator to register a sentry integration.

    :param object args: integration class constructor arguments.
    :param object kwargs: integration class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           integration with the same name, replace it with
                           the new one. otherwise raise an error.
                           defaults to False.

    :raises InvalidSentryIntegrationTypeError: invalid sentry integration type error.
    :raises DuplicatedSentryIntegrationError: duplicated sentry integration error.

    :returns: integration class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available sentry integrations.

        :param type cls: integration class.

        :returns: integration class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        sentry_services.register_integration(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
sentry manager module.
"""

import sentry_sdk

import pyrin.configuration.services as config_services

from pyrin.caching.mixin.decorators import fast_cache
from pyrin.caching.mixin.typed import TypedCacheMixin
from pyrin.core.structs import Manager, Context
from pyrin.logging.enumerations import LogLevelEnum, LogLevelIntEnum
from pyrin.logging.sentry import SentryPackage
from pyrin.logging.sentry.interface import AbstractSentryIntegration
from pyrin.utils.custom_print import print_warning
from pyrin.logging.sentry.exceptions import InvalidSentryIntegrationTypeError, \
    DuplicatedSentryIntegrationError, SentryIntegrationNotFoundError


class SentryManager(Manager, TypedCacheMixin):
    """
    sentry manager class.
    """

    package_class = SentryPackage

    def __init__(self):
        """
        initializes an instance of SentryManager.
        """

        super().__init__()

        # a dict containing different integrations in the form of:
        # {str name: AbstractSentryIntegration integration}
        self._integrations = Context()

    def _get_integration(self, name):
        """
        gets the registered integration with given name.

        it raises an error if no integration found for given name.

        :param str name: name of the integration to be get.

        :raises SentryIntegrationNotFoundError: sentry integration not found error.

        :rtype: AbstractSentryIntegration
        """

        if name not in self._integrations:
            raise SentryIntegrationNotFoundError('Sentry integration [{name}] does '
                                                 'not exist.'.format(name=name))

        return self._integrations.get(name)

    @fast_cache
    def _is_logging_integration_enabled(self):
        """
        gets a value indicating that logging integration is enabled.

        :rtype: bool
        """

        default_integrations_enabled = config_services.get_active('sentry',
                                                                  'default_integrations')

        logging_enabled = config_services.get('sentry', 'logging', 'enable')
        logging_event_level = config_services.get('sentry', 'logging', 'event_level')

        return (default_integrations_enabled is True and logging_enabled is False) or \
               (logging_enabled is True and logging_event_level in (LogLevelEnum.ERROR,
                                                                    LogLevelEnum.WARNING,
                                                                    LogLevelEnum.INFO,
                                                                    LogLevelEnum.DEBUG,
                                                                    LogLevelIntEnum.ERROR,
                                                                    LogLevelIntEnum.WARNING,
                                                                    LogLevelIntEnum.INFO,
                                                                    LogLevelIntEnum.DEBUG))

    def register_integration(self, instance, **options):
        """
        registers a new integration or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding an integration which is already registered.

        :param AbstractSentryIntegration instance: integration instance to be registered.

        :keyword bool replace: specifies that if there is another registered
                               integration with the same name, replace it with the
                               new one, otherwise raise an error. defaults to False.

        :raises InvalidSentryIntegrationTypeError: invalid sentry integration type error.
        :raises DuplicatedSentryIntegrationError: duplicated sentry integration error.
        """

        if not isinstance(instance, AbstractSentryIntegration):
            raise InvalidSentryIntegrationTypeError('Input parameter [{instance}] is '
                                                    'not an instance of [{base}].'
                                                    .format(instance=instance,
                                                            base=AbstractSentryIntegration))

        if instance.get_name() in self._integrations:
            old_instance = self._get_integration(instance.get_name())
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedSentryIntegrationError('There is another registered sentry '
                                                       'integration [{old}] with name [{name}] '
                                                       'but "replace" option is not set, so '
                                                       'integration [{instance}] could not be '
                                                       'registered.'
                                                       .format(old=old_instance,
                                                               name=instance.get_name(),
                                                               instance=instance))

            print_warning('Sentry integration [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._integrations[instance.get_name()] = instance

    def configure(self, **options):
        """
        configures sentry client and all enabled integrations.
        """

        integrations = []
        for name, instance in self._integrations.items():
            instance.configure(integrations)

        configs = config_services.get_active_section('sentry')
        sentry_sdk.init(**configs, integrations=integrations)

    def try_report_exception(self, error, **options):
        """
        tries to report exception to sentry server.

        if logging integration is not enabled to send it automatically.

        :param Exception error: exception instance that has been occurred.
        """

        if self._is_logging_integration_enabled() is False:
            sentry_sdk.capture_exception(error)



# -*- coding: utf-8 -*-
"""
orm types base module.
"""

from abc import abstractmethod

from sqlalchemy import TypeDecorator

from pyrin.core.exceptions import CoreNotImplementedError


class CoreCustomType(TypeDecorator):
    """
    core custom type class.
    all application custom types must be subclassed from this type.
    """

    @abstractmethod
    def load_dialect_impl(self, dialect):
        """
        returns a `TypeEngine` object corresponding to a dialect.

        :param Dialect dialect: the dialect in use.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: TypeEngine
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def process_bind_param(self, value, dialect):
        """
        receive a bound parameter value to be converted.

        :param TypeEngine value: data to operate upon. it could be `None`.
        :param Dialect dialect: the dialect in use.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def process_result_value(self, value, dialect):
        """
        receive a result-row column value to be converted.

        :param str value: data to operate upon. it could be `None`.
        :param Dialect dialect: the dialect in use.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def compare_against_backend(self, dialect, conn_type):
        """
        returns True if this type is the same as the given database type,
        or None to allow the default implementation to compare these
        types. a return value of False means the given type does not
        match this type.

        :param Dialect dialect: the dialect in use.
        :param TypeEngine conn_type: type of the returned value from database.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()
